from ..value_set import ValueSet


class DtapVaccine(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts for immunizations with diphtheria, tetanus, and pertussis (DTaP) vaccines.

    **Data Element Scope:** This value set may use a model element related to Immunization.

    **Inclusion Criteria:** Includes concepts that represent a diphtheria, tetanus, and pertussis (DTaP) vaccine.

    **Exclusion Criteria:** No exclusions.

    ** Used in:** CMS117v10
    """

    VALUE_SET_NAME = 'DTaP Vaccine'
    OID = '2.16.840.1.113883.3.464.1003.196.12.1214'
    DEFINITION_VERSION = '20170504'
    EXPANSION_VERSION = 'eCQM Update 2021-05-06'

    CVX = {
        '20',  # diphtheria, tetanus toxoids and acellular pertussis vaccine
        '50',  # DTaP-Haemophilus influenzae type b conjugate vaccine
        '106',  # diphtheria, tetanus toxoids and acellular pertussis vaccine, 5 pertussis antigens
        '107',  # diphtheria, tetanus toxoids and acellular pertussis vaccine, unspecified formulation
        '110',  # DTaP-hepatitis B and poliovirus vaccine
        '120',  # diphtheria, tetanus toxoids and acellular pertussis vaccine, Haemophilus influenzae type b conjugate, and poliovirus vaccine, inactivated (DTaP-Hib-IPV)
    }


class HepatitisAVaccine(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts for immunizations with Hepatitis A vaccines.

    **Data Element Scope:** This value set may use a model element related to Immunization.

    **Inclusion Criteria:** Includes concepts that represent a Hepatitis A vaccine.

    **Exclusion Criteria:** No exclusions.

    ** Used in:** CMS117v10
    """

    VALUE_SET_NAME = 'Hepatitis A Vaccine'
    OID = '2.16.840.1.113883.3.464.1003.196.12.1215'
    DEFINITION_VERSION = '20170504'
    EXPANSION_VERSION = 'eCQM Update 2021-05-06'

    CVX = {
        '31',  # hepatitis A vaccine, pediatric dosage, unspecified formulation
        '83',  # hepatitis A vaccine, pediatric/adolescent dosage, 2 dose schedule
        '85',  # hepatitis A vaccine, unspecified formulation
    }


class HepatitisBVaccine(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts for immunizations with Hepatitis B vaccines.

    **Data Element Scope:** This value set may use a model element related to Immunization.

    **Inclusion Criteria:** Includes concepts that represent a Hepatitis B vaccine.

    **Exclusion Criteria:** No exclusions.

    ** Used in:** CMS117v10
    """

    VALUE_SET_NAME = 'Hepatitis B Vaccine'
    OID = '2.16.840.1.113883.3.464.1003.196.12.1216'
    DEFINITION_VERSION = '20170504'
    EXPANSION_VERSION = 'eCQM Update 2021-05-06'

    CVX = {
        '08',  # hepatitis B vaccine, pediatric or pediatric/adolescent dosage
        '44',  # hepatitis B vaccine, dialysis patient dosage
        '45',  # hepatitis B vaccine, unspecified formulation
        '51',  # Haemophilus influenzae type b conjugate and Hepatitis B vaccine
        '110',  # DTaP-hepatitis B and poliovirus vaccine
    }


class HibVaccine3DoseSchedule(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts for immunizations with 3-dose haemophilus influenzae type b (Hib) vaccines.

    **Data Element Scope:** This value set may use a model element related to Immunization.

    **Inclusion Criteria:** Includes concepts that represent a 3-dose haemophilus influenzae type b (Hib) vaccine.

    **Exclusion Criteria:** No exclusions.

    ** Used in:** CMS117v10
    """

    VALUE_SET_NAME = 'Hib Vaccine (3 dose schedule)'
    OID = '2.16.840.1.113883.3.464.1003.110.12.1083'
    DEFINITION_VERSION = '20190315'
    EXPANSION_VERSION = 'eCQM Update 2021-05-06'

    CVX = {
        '49',  # Haemophilus influenzae type b vaccine, PRP-OMP conjugate
    }


class HibVaccine4DoseSchedule(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts for immunizations with 4-dose haemophilus influenzae type b (Hib) vaccines.

    **Data Element Scope:** This value set may use a model element related to Immunization.

    **Inclusion Criteria:** Includes concepts that represent a 4-dose haemophilus influenzae type b (Hib) vaccine.

    **Exclusion Criteria:** No exclusions.

    ** Used in:** CMS117v10
    """

    VALUE_SET_NAME = 'Hib Vaccine (4 dose schedule)'
    OID = '2.16.840.1.113883.3.464.1003.110.12.1085'
    DEFINITION_VERSION = '20190315'
    EXPANSION_VERSION = 'eCQM Update 2021-05-06'

    CVX = {
        '48',  # Haemophilus influenzae type b vaccine, PRP-T conjugate
        '50',  # DTaP-Haemophilus influenzae type b conjugate vaccine
        '51',  # Haemophilus influenzae type b conjugate and Hepatitis B vaccine
        '120',  # diphtheria, tetanus toxoids and acellular pertussis vaccine, Haemophilus influenzae type b conjugate, and poliovirus vaccine, inactivated (DTaP-Hib-IPV)
        '148',  # Meningococcal Groups C and Y and Haemophilus b Tetanus Toxoid Conjugate Vaccine
    }


class InactivatedPolioVaccineIpv(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts for immunizations with inactivated polio vaccines (IPV).

    **Data Element Scope:** This value set may use a model element related to Immunization.

    **Inclusion Criteria:** Includes concepts that represent an inactivated polio vaccine (IPV).

    **Exclusion Criteria:** No exclusions.

    ** Used in:** CMS117v10
    """

    VALUE_SET_NAME = 'Inactivated Polio Vaccine (IPV)'
    OID = '2.16.840.1.113883.3.464.1003.196.12.1219'
    DEFINITION_VERSION = '20170504'
    EXPANSION_VERSION = 'eCQM Update 2021-05-06'

    CVX = {
        '10',  # poliovirus vaccine, inactivated
        '89',  # poliovirus vaccine, unspecified formulation
        '110',  # DTaP-hepatitis B and poliovirus vaccine
        '120',  # diphtheria, tetanus toxoids and acellular pertussis vaccine, Haemophilus influenzae type b conjugate, and poliovirus vaccine, inactivated (DTaP-Hib-IPV)
    }


class InfluenzaVaccine(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts for immunizations with influenza vaccines.

    **Data Element Scope:** This value set may use a model element related to Immunization.

    **Inclusion Criteria:** Includes concepts that represent an influenza vaccine.

    **Exclusion Criteria:** No exclusions.

    ** Used in:** CMS147v11, CMS117v10
    """

    VALUE_SET_NAME = 'Influenza Vaccine'
    OID = '2.16.840.1.113883.3.526.3.1254'
    DEFINITION_VERSION = '20170908'
    EXPANSION_VERSION = 'eCQM Update 2021-05-06'

    CVX = {
        '88',  # influenza virus vaccine, unspecified formulation
        '135',  # influenza, high dose seasonal, preservative-free
        '140',  # Influenza, seasonal, injectable, preservative free
        '141',  # Influenza, seasonal, injectable
        '144',  # seasonal influenza, intradermal, preservative free
        '149',  # influenza, live, intranasal, quadrivalent
        '150',  # Influenza, injectable, quadrivalent, preservative free
        '153',  # Influenza, injectable, Madin Darby Canine Kidney, preservative free
        '155',  # Seasonal, trivalent, recombinant, injectable influenza vaccine, preservative free
        '158',  # influenza, injectable, quadrivalent, contains preservative
        '161',  # Influenza, injectable,quadrivalent, preservative free, pediatric
        '166',  # influenza, intradermal, quadrivalent, preservative free, injectable
        '168',  # Seasonal trivalent influenza vaccine, adjuvanted, preservative free
        '171',  # Influenza, injectable, Madin Darby Canine Kidney, preservative free, quadrivalent
        '185',  # Seasonal, quadrivalent, recombinant, injectable influenza vaccine, preservative free
        '186',  # Influenza, injectable, Madin Darby Canine Kidney,  quadrivalent with preservative
        '197',  # influenza, high-dose seasonal, quadrivalent, .7mL dose, preservative free
        '205',  # influenza, seasonal vaccine, quadrivalent, adjuvanted, .5mL dose, preservative free
    }


class InfluenzaVirusLaivImmunization(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent live attenuated influenza vaccine.

    **Data Element Scope:** This value set may use a model element related to Immunization.

    **Inclusion Criteria:** Includes concepts that represent a live attenuated influenza vaccine (quadrivalent and trivalent).

    **Exclusion Criteria:** Excludes concepts that represent the recombinant or inactivated influenza vaccine.

    ** Used in:** CMS117v10
    """

    VALUE_SET_NAME = 'Influenza Virus LAIV Immunization'
    OID = '2.16.840.1.113883.3.464.1003.110.12.1087'
    DEFINITION_VERSION = '20210224'
    EXPANSION_VERSION = 'eCQM Update 2021-05-06'

    CVX = {
        '111',  # influenza virus vaccine, live, attenuated, for intranasal use
        '149',  # influenza, live, intranasal, quadrivalent
    }


class Measles_MumpsAndRubellaMmrVaccine(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts for immunizations with measles, mumps and rubella (MMR) vaccines.

    **Data Element Scope:** This value set may use a model element related to Immunization.

    **Inclusion Criteria:** Includes concepts that represent a measles, mumps and rubella (MMR) vaccine.

    **Exclusion Criteria:** No exclusions.

    ** Used in:** CMS117v10
    """

    VALUE_SET_NAME = 'Measles, Mumps and Rubella (MMR) Vaccine'
    OID = '2.16.840.1.113883.3.464.1003.196.12.1224'
    DEFINITION_VERSION = '20170504'
    EXPANSION_VERSION = 'eCQM Update 2021-05-06'

    CVX = {
        '03',  # measles, mumps and rubella virus vaccine
        '94',  # measles, mumps, rubella, and varicella virus vaccine
    }


class PneumococcalConjugateVaccine(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts for immunizations with pneumococcal conjugate vaccines.

    **Data Element Scope:** This value set may use a model element related to Immunization.

    **Inclusion Criteria:** Includes concepts that represent a pneumococcal conjugate vaccine.

    **Exclusion Criteria:** No exclusions.

    ** Used in:** CMS117v10
    """

    VALUE_SET_NAME = 'Pneumococcal Conjugate Vaccine'
    OID = '2.16.840.1.113883.3.464.1003.196.12.1221'
    DEFINITION_VERSION = '20170504'
    EXPANSION_VERSION = 'eCQM Update 2021-05-06'

    CVX = {
        '133',  # pneumococcal conjugate vaccine, 13 valent
        '152',  # Pneumococcal Conjugate, unspecified formulation
    }


class RotavirusVaccine3DoseSchedule(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts for immunizations with 3-dose rotavirus vaccines.

    **Data Element Scope:** This value set may use a model element related to Immunization.

    **Inclusion Criteria:** Includes concepts that represent a 3-dose rotavirus vaccine.

    **Exclusion Criteria:** No exclusions.

    ** Used in:** CMS117v10
    """

    VALUE_SET_NAME = 'Rotavirus Vaccine (3 dose schedule)'
    OID = '2.16.840.1.113883.3.464.1003.196.12.1223'
    DEFINITION_VERSION = '20170504'
    EXPANSION_VERSION = 'eCQM Update 2021-05-06'

    CVX = {
        '116',  # rotavirus, live, pentavalent vaccine
        '122',  # rotavirus vaccine, unspecified formulation
    }


class VaricellaZosterVaccineVzv(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts for immunizations with varicella zoster vaccines.

    **Data Element Scope:** This value set may use a model element related to Immunization.

    **Inclusion Criteria:** Includes concepts that represent a varicella zoster vaccine.

    **Exclusion Criteria:** No exclusions.

    ** Used in:** CMS117v10
    """

    VALUE_SET_NAME = 'Varicella Zoster Vaccine (VZV)'
    OID = '2.16.840.1.113883.3.464.1003.196.12.1170'
    DEFINITION_VERSION = '20170504'
    EXPANSION_VERSION = 'eCQM Update 2021-05-06'

    CVX = {
        '21',  # varicella virus vaccine
        '94',  # measles, mumps, rubella, and varicella virus vaccine
    }


class PneumococcalPolysaccharide23Vaccine(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts for immunizations with pneumococcal polysaccharide vaccines.

    **Data Element Scope:** This value set may use a model element related to Immunization.

    **Inclusion Criteria:** Includes concepts that represent a pneumococcal polysaccharide 23-valent vaccine.

    **Exclusion Criteria:** Excludes concepts that represent the pneumococcal conjugate 7-valent and 13-valent vaccines.

    ** Used in:** CMS127v10
    """

    VALUE_SET_NAME = 'Pneumococcal Polysaccharide 23 Vaccine'
    OID = '2.16.840.1.113883.3.464.1003.110.12.1089'
    DEFINITION_VERSION = '20210224'
    EXPANSION_VERSION = 'eCQM Update 2021-05-06'

    CVX = {
        '33',  # pneumococcal polysaccharide vaccine, 23 valent
        '109',  # pneumococcal vaccine, unspecified formulation
    }




from protocols import events
from protocols.protocol import ClinicalQualityMeasure
from protocols.value_set.v2021 import (
    BmiPercentile, CounselingForNutrition, CounselingForPhysicalActivity, EncounterInpatient,
    Ethnicity, Height, HomeHealthcareServices, HospiceCareAmbulatory, OfficeVisit,
    OncAdministrativeSex, Payer, Pregnancy, PreventiveCareEstablishedOfficeVisit0To17,
    PreventiveCareServicesGroupCounseling, PreventiveCareServicesIndividualCounseling,
    PreventiveCareServicesInitialOfficeVisit0To17, Race, Weight)


class ClinicalQualityMeasure155v9(ClinicalQualityMeasure):
    """
    Weight Assessment and Counseling for Nutrition and Physical Activity for Children and
    Adolescents

    Description: Percentage of patients 3-17 years of age who had an outpatient visit with a
    Primary Care Physician (PCP) or Obstetrician/Gynecologist (OB/GYN) and who had evidence of the
    following during the measurement period. Three rates are reported.

     - Percentage of patients with height, weight, and body mass index (BMI) percentile
    documentation
     - Percentage of patients with counseling for nutrition
     - Percentage of patients with counseling for physical activity

    Definition: None

    Rationale: Over the last four decades, childhood obesity has more than tripled in children and
    adolescents 2 to 19 years of age (from a rate of approximately 5 percent to 18.5 percent)
    (Fryar, Carroll, & Ogden, 2014; Hales et al., 2017). Non-Hispanic black and Hispanic youth are
    more likely to be obese than their non-Hispanic white and non-Hispanic Asian counterparts. In
    2015-2016, approximately 22 percent of non-Hispanic black and 26 percent of Hispanic youth were
    obese compared to approximately 14 percent of non-Hispanic white and 11 percent of non-Hispanic
    Asian youth (Hales et al., 2017).

    Childhood obesity has both immediate and long-term effects on health and well-being. Children
    who are obese have higher rates of physical health conditions, such as risk factors for
    cardiovascular disease (like high blood pressure and high cholesterol), type 2 diabetes,
    asthma, sleep apnea, and joint problems. There is also a correlation between childhood obesity
    and mental health conditions, such as anxiety and depression (Centers for Disease Control and
    Prevention, 2016). In addition, children who are obese are more likely to be obese as adults
    and are therefore at risk for adult health problems, such as heart disease, type 2 diabetes,
    and several types of cancer (Centers for Disease Control and Prevention, 2016).

    The direct medical costs associated with childhood obesity total about $19,000 per child,
    contributing to the $14 billion spent on care related to childhood obesity in the United States
    (Finkelstein, Graham, & Malhotra, 2014).

    Because obesity can become a lifelong health issue, it is important to screen for obesity in
    children and adolescents, and to provide interventions that promote weight loss (U.S.
    Preventive Services Task Force, 2017).

    Guidance: The visit must be performed by a PCP or OB/GYN.
    Because BMI norms for youth vary with age and sex, this measure evaluates whether BMI
    percentile, rather than an absolute BMI value, is assessed.

    This eCQM is a patient-based measure.

    This version of the eCQM uses QDM version 5.5.  Please refer to the eCQI resource center
    (https://ecqi.healthit.gov/qdm) for more information on the QDM.

    More information: https://ecqi.healthit.gov/ecqm/measures/cms155v9
    """

    title = 'Weight Assessment and Counseling for Nutrition and Physical Activity for Children and Adolescents'

    identifiers = ['CMS155v9']

    types = ['CQM']

    responds_to_event_types = [
        events.HEALTH_MAINTENANCE,
    ]

    authors = [
        'National Committee for Quality Assurance',
    ]

    references = [
        'Centers for Disease Control and Prevention. (2016). Childhood obesity causes & consequences. Retrieved from https://www.cdc.gov/obesity/childhood/causes.html',
        'Finkelstein, E. A., Graham, W. C. K., & Malhotra, R. (2014). Lifetime direct medical costs of childhood obesity. Pediatrics, 133(5), 854-862.',
        'Fryar, C. D., Carroll, M. D., & Ogden, C. L. (2014). Prevalence of overweight and obesity among children and adolescents: United States, 1963-1965 through 2011-2012. Health E-Stats. Retrieved from https://www.cdc.gov/nchs/data/hestat/obesity_child_11_12/obesity_child_11_12.htm',
        'Hagan, J. F., Shaw, J. S., & Duncan, P. M. (eds.). (2017). Bright futures: Guidelines for health supervision of infants, children, and adolescents, 4th ed. Elk Grove Village, IL: American Academy of Pediatrics.',
        'Hales, C.M., Carroll, M.D., Fryar C.D., et al. (2017). Prevalence of obesity among adults and youth: United States, 2015-2016. NCHS Data Brief. Retrieved from https://www.cdc.gov/nchs/data/databriefs/db288.pdf',
        'U.S. Preventive Services Task Force. (2017). Screening and interventions for overweight in children and adolescents: Recommendation statement. Rockville, MD: Agency for Healthcare Research and Quality. https://www.uspreventiveservicestaskforce.org/Page/Document/UpdateSummaryFinal/obesity-in-children-and-adolescents-screening1?ds=1&s=obesity',
    ]

    funding_source = ''

    def in_initial_population(self):
        """
        Initial population: Patients 3-17 years of age with at least one outpatient visit with a
        primary care physician (PCP) or an obstetrician/gynecologist (OB/GYN) during the
        measurement period
        """
        pass

    def in_denominator(self):
        """
        Denominator: Equals Initial Population

        Exclusions: Patients who have a diagnosis of pregnancy during the measurement period.

        Exclude patients whose hospice care overlaps the measurement period.

        Exceptions: None
        """
        return self.in_initial_population()

    def in_numerator(self):
        """
        Numerator: Numerator 1: Patients who had a height, weight and body mass index (BMI)
        percentile recorded during the measurement period
        Numerator 2: Patients who had counseling for nutrition during the measurement period
        Numerator 3: Patients who had counseling for physical activity during the measurement
        period

        Exclusions: Not Applicable
        """
        pass


    def compute_results(self):
        """
        Clinical recommendation: U.S. Preventive Services Task Force (2017) - The task force
        recommends that clinicians screen for obesity in children and adolescents 6 years and older
        and offer or refer them to comprehensive, intensive behavioral interventions to promote
        improvements in weight status. (B recommendation)

        American Academy of Pediatrics – Bright Futures (Hagan, Shaw, & Duncan, 2017) -

        - Plot and assess BMI percentiles routinely for early recognition of overweight and
        obesity.
        - Assess barriers to healthy eating and physical activity.
        - Provide anticipatory guidance for nutrition and physical activity.
        """
        pass



from datetime import datetime, timedelta

import arrow

from canvas_workflow_kit.patient_recordset import ProtocolOverrideRecordSet
from canvas_workflow_kit.protocol import (
    ClinicalQualityMeasure,
    Protocol,
    ProtocolResult,
    ProtocolSource,
    get_protocols,
    get_subclasses,
    protocols_for_patient
)
from canvas_workflow_kit.recommendation import Recommendation, PlanRecommendation
from canvas_workflow_kit.tests import get_protocols_unittest
from canvas_workflow_kit.timeframe import Timeframe

from .base import SDKBaseTest


LOADED_PROTOCOLS = get_protocols(get_protocols_unittest)

class TestProtocolResult(SDKBaseTest):

    def test___init__(self):
        tested = ProtocolResult()
        self.assertEqual([], tested.recommendations)
        self.assertEqual('not_applicable', tested.status)
        self.assertEqual([], tested.narratives)
        self.assertIsNone(tested.due_in)
        self.assertEqual(30, tested.days_of_notice)
        self.assertIsNone(tested.next_review)

    def test_add_recommendation(self):
        tested = ProtocolResult()
        self.assertEqual([], tested.recommendations)
        tested.add_recommendation(
            Recommendation('KEY-A', 7, 'ACT', 'reco A', 'this is the reco A'))
        tested.add_recommendation(
            Recommendation('KEY-B', 3, 'BAC', 'reco B', 'this is the reco B'))
        tested.add_recommendation(
            PlanRecommendation('KEY-C', 4, 'BAC', 'button', 'reco C', 'this is the reco C'))
        self.assertEqual(3, len(tested.recommendations))
        for recommendation in tested.recommendations:
            self.assertIn(recommendation.title, ['reco A', 'reco B', 'reco C'])

    def test_add_recommendation_shorthand(self):
        tested = ProtocolResult()
        self.assertEqual([], tested.recommendations)
        tested.add_recommendation(
            'KEY-A', 7, 'ACT', 'reco A', 'this is the reco A')
        tested.add_recommendation(
            key='KEY-B', rank=3, button='BAC', title='reco B', narrative='this is the reco B')
        self.assertEqual(2, len(tested.recommendations))
        for recommendation in tested.recommendations:
            self.assertIn(recommendation.title, ['reco A', 'reco B'])

    def test_add_recommendation_shorthand_custom(self):
        tested = ProtocolResult()
        self.assertEqual([], tested.recommendations)
        tested.add_plan_recommendation('KEY-A', 7, 'ACT', 'reco A', 'this is the reco A')
        tested.add_vital_sign_recommendation('KEY-B', 3, 'BAC', 'reco B', 'this is the reco B')
        self.assertEqual(2, len(tested.recommendations))

        self.assertEqual(tested.recommendations[0].__class__.__name__, 'PlanRecommendation')
        self.assertEqual(tested.recommendations[1].__class__.__name__, 'VitalSignRecommendation')

    def test_add_narrative(self):
        tested = ProtocolResult()
        self.assertEqual([], tested.narratives)
        tested.add_narrative('narrative A')
        tested.add_narrative('narrative B')
        self.assertEqual(2, len(tested.narratives))
        for narrative in tested.narratives:
            self.assertIn(narrative, ['narrative A', 'narrative B'])

    def test_narrative(self):
        tested = ProtocolResult()
        self.assertEqual('', tested.narrative)
        tested.add_narrative('narrative A')
        tested.add_narrative('narrative B')
        expected = 'narrative A\nnarrative B'
        self.assertEqual(expected, tested.narrative)


class TestProtocolSource(SDKBaseTest):

    def test___init__(self):
        tested = ProtocolSource('Some text', 'https://not.known.com/utests')
        self.assertIsInstance(tested, ProtocolSource)
        self.assertEqual('Some text', tested.text)
        self.assertEqual('https://not.known.com/utests', tested.url)


class TestProtocol(SDKBaseTest):

    def test___init__(self):
        patient = self.load_patient('full')
        tested = Protocol(patient)
        self.assertIsInstance(tested, Protocol)
        self.assertIs(patient, tested.patient)
        self.assertEqual([], tested._meta.compute_on_change_types)
        self.assertEqual([], tested._meta.responds_to_event_types)
        self.assertEqual([], tested._meta.identifiers)
        self.assertEqual('', tested._meta.title)
        self.assertEqual('', tested._meta.description)
        self.assertEqual('', tested._meta.information)

    def test_protocol_key(self):
        patient = self.load_patient('full')
        tested = Protocol(patient)
        expected = 'Protocol'
        result = tested.protocol_key()
        self.assertEqual(expected, result)

        tested = TestProto(patient)
        expected = 'TestProto'
        result = tested.protocol_key()
        self.assertEqual(expected, result)

    def test_protocol_title(self):
        patient = self.load_patient('full')
        tested = Protocol(patient)
        expected = ''
        result = tested._meta.title
        self.assertEqual(expected, result)

        tested = TestProto(patient)
        expected = 'Title TestProtocol'
        result = tested._meta.title
        self.assertEqual(expected, result)

    def test_references(self):
        patient = self.load_patient('full')
        tested = Protocol(patient)

        tested = TestProto(patient)
        expected = [
            'Small title #1. www.unittests.med/canvas?id=1',
            'Small title #2. http://www.unittests.med/canvas?id=2',
            'Small title #3. https://www.unittests.med/canvas?id=3',
            'Small title #4. WWW.unittests.med/canvas?id=4',
            'Not valid #5  www.unittests.med/canvas?id=5 because of this',
        ]
        self.assertEqual(expected, tested._meta.references)

    def test_relative_float(self):
        tests = [
            ['123.45', 123.45],
            ['≤123.45', 123.45],
            ['≥123.45', 123.45],
            ['>=123.45', 123.45],
            ['<=123.45', 123.45],
            ['>123.45', 123.450001],
            ['<123.45', 123.449999],
            ['-123.45', -123.45],
            ['=123.45', 123.45],
            ['=', 0],
            ['', 0],
            ['abcd', 0],
        ]
        for value, expected in tests:
            result = Protocol.relative_float(value)
            self.assertEqual(expected, result, value)

    def test_sources(self):
        patient = self.load_patient('full')
        tested = TestProto(patient)

        expected = [
            'www.unittests.med/canvas?id=1',
            'http://www.unittests.med/canvas?id=2',
            'https://www.unittests.med/canvas?id=3',
            'WWW.unittests.med/canvas?id=4',
            None,
        ]
        result = tested.sources()
        self.assertEqual(len(expected), len(result))
        for src in result:
            self.assertIsInstance(src, ProtocolSource)
            self.assertEqual(1, expected.count(src.url))

    def test_status(self):
        patient = self.load_patient('full')
        tested = Protocol(patient)
        with self.assertRaises(NotImplementedError):
            tested.status()

        tested = TestProto(patient)
        expected = 'current status'
        self.assertEqual(expected, tested.status())

    def test_snoozed(self):
        patient = self.load_patient('full')
        patient.protocol_overrides = ProtocolOverrideRecordSet([
            {
                'protocolKey': 'TEST001v2',
                'adjustment': {
                    'reference': '2018-10-01T00:00:00Z',
                    'cycleDays': 19
                },
                'snooze': {
                    'reference': arrow.now(),
                    'snoozedDays': 22
                },
                'modified': '2018-10-02T21:50:09.661490Z'
            },
            {
                'protocolKey': 'TEST003v2',
                'adjustment': None,
                'snooze': {
                    'reference': '2018-09-15T00:00:00Z',
                    'snoozedDays': 12
                },
                'modified': '2018-09-15T21:50:09.661490Z'
            },
            {
                'protocolKey': 'TEST004v1',
                'adjustment': {
                    'reference': '2018-08-15T00:00:00Z',
                    'cycleDays': 60
                },
                'snooze': None,
                'modified': '2018-08-23T21:50:09.661490Z'
            }])  # yapf: disable

        # no protocol override --> false
        tested = Protocol(patient=patient)
        tested.identifiers = ['TEST005v1']
        self.assertFalse(tested.snoozed())

        # protocol override without snooze --> false
        tested = Protocol(patient=patient)
        tested.identifiers = ['TEST004v1']
        self.assertFalse(tested.snoozed())

        # protocol override with past snooze --> false
        tested = Protocol(patient=patient)
        tested.identifiers = ['TEST003v2']
        self.assertFalse(tested.snoozed())

        # protocol override with snooze --> true
        tested = Protocol(patient=patient)
        tested.protocol_key = lambda: 'TEST001v2'
        tested.identifiers = ['TEST001v2']
        self.assertTrue(tested.snoozed())

        # protocol override switch off --> true
        patient.protocol_overrides = ProtocolOverrideRecordSet([{
            'protocolKey': '*',
            'adjustment': None,
            'snooze': None,
            'modified': '2018-08-23T21:50:09.661490Z'
        }])
        tested = Protocol(patient=patient)
        tested.identifiers = ['TEST005v2']
        self.assertTrue(tested.snoozed())

    def test_enabled(self):
        patient = self.load_patient('full')
        tested = TestProto(patient)
        self.assertTrue(tested.enabled())

        tested = TestPFalse(patient)
        self.assertFalse(tested.enabled())


class TestClinicalQualityMeasure(SDKBaseTest):

    def test___init__(self):
        patient = self.load_patient('full')

        # no change types, time frame provided
        tested = CQMImplementation(patient=patient)

        self.assertEqual(['HEALTH_MAINTENANCE'], tested._meta.responds_to_event_types)
        self.assertIsNone(tested.results())

        result = tested.timeframe
        self.assertIsInstance(result, Timeframe)
        fmt = '%Y-%m-%d %H:%M'

        last_year = {
            'year': datetime.utcnow().year - 1,
            'month': datetime.utcnow().month,
            'day': datetime.utcnow().day,
            'hour': datetime.utcnow().hour,
            'minute': datetime.utcnow().minute,
        }

        self.assertEqual(datetime(**last_year).strftime(fmt), tested.timeframe.start.strftime(fmt))
        self.assertEqual(datetime.utcnow().strftime(fmt), tested.timeframe.end.strftime(fmt))

        self.assertEqual(datetime.utcnow().strftime(fmt), tested.now.strftime(fmt))
        self.assertEqual('report', tested.context)

        # context provided
        tests = {
            'report': 'report',
            'guidance': 'guidance',
            'nope': 'report',
        }
        for context, expected in tests.items():
            tested = CQMImplementation(patient=patient, context=context)
            self.assertEqual(expected, tested.context)

        # time frame provided
        start = arrow.get('2018-08-17 01:23:45')
        end = arrow.get('2018-08-17 12:34:56')
        timeframe = Timeframe(start, end)
        tested = ClinicalQualityMeasure(patient=patient, timeframe=timeframe)
        result = tested.timeframe
        self.assertIsInstance(result, Timeframe)
        self.assertIs(start, result.start)
        self.assertIs(end, result.end)

        self.assertEqual(datetime.utcnow().strftime(fmt), tested.now.strftime(fmt))

        # now provided
        tested = ClinicalQualityMeasure(patient=patient, now=arrow.get('2018-10-30 12:34'))
        self.assertEqual('2018-10-30 12:34', tested.now.strftime(fmt))

        # adjustment
        patient.protocol_overrides = ProtocolOverrideRecordSet([{
            'protocolKey': 'TestCQM',
            'adjustment': {
                'reference': '2018-02-01T00:00:00Z',
                'cycleDays': 180
            },
            'snooze': None,
            'modified': '2018-03-15T12:32:54Z',
        }])  # yapf: disable
        tested = TestCQM(patient=patient, now=arrow.get('2018-10-30 12:34'))
        result = tested.timeframe
        self.assertEqual('2018-05-03 12:34', result.start.strftime(fmt))
        self.assertEqual('2018-10-30 12:34', result.end.strftime(fmt))

        # change types provided
        # + empty -> instance is computable
        tested = CQMImplementation(patient=patient, change_types=[])

        self.assertEqual(['HEALTH_MAINTENANCE'], tested._meta.responds_to_event_types)
        self.assertIsNone(tested.results())

        # + with changes + instance with empty impacting list -> instance is computable
        tested = CQMImplementation(
            patient=patient, change_types=[
                'change_A',
                'change_B',
            ])

        self.assertEqual(['HEALTH_MAINTENANCE'], tested._meta.responds_to_event_types)
        self.assertIsNone(tested.results())

        # + with changes + instance impacted by one change -> instance is computable
        class UnitTestCQMImpactedByA(CQMImplementation):
            class Meta:
                compute_on_change_types = [
                    'change_A',
                    'change_C',
                ]

        tested = UnitTestCQMImpactedByA(
            patient=patient, change_types=[
                'change_A',
                'change_B',
            ])

        self.assertEqual(['HEALTH_MAINTENANCE'], tested._meta.responds_to_event_types)
        self.assertIsNone(tested.results())

        # + with changes + instance impacted by other changes -> instance is not computable
        tested = UnitTestCQMImpactedByA(
            patient=patient, change_types=[
                'change_D',
                'change_B',
            ])

        self.assertEqual(['HEALTH_MAINTENANCE'], tested._meta.responds_to_event_types)
        self.assertIsNotNone(tested.results())
        self.assertEqual('not_relevant', tested.status())

    def test_impacted_by_changes(self):
        patient = self.load_patient('full')
        tested = CQMImplementation(patient=patient, change_types=[])

        # no change types -> True
        result = tested.impacted_by_changes([])
        self.assertTrue(result)
        # change types + no impacting changes -> True
        result = tested.impacted_by_changes([
            'change_A',
            'change_B',
            'change_C',
        ])
        self.assertTrue(result)
        # change types + impacting changes in the changes -> True
        tested._meta.compute_on_change_types = [
            'change_X',
            'change_B',
            'change_C',
        ]
        result = tested.impacted_by_changes([
            'change_A',
            'change_B',
            'change_C',
        ])
        self.assertTrue(result)
        result = tested.impacted_by_changes([
            'change_A',
            'change_B',
        ])
        self.assertTrue(result)
        result = tested.impacted_by_changes([
            'change_C',
        ])
        self.assertTrue(result)
        # change types + impacting changes Not in the changes -> False
        result = tested.impacted_by_changes([
            'change_U',
            'change_V',
            'change_W',
        ])
        self.assertFalse(result)

    def test_period_adjustment(self):
        patient = self.load_patient('full')

        tested = TestCQM(patient=patient, now=arrow.get('2018-10-30 12:34'))
        result = tested.period_adjustment
        self.assertIsNone(result)

        patient.protocol_overrides = ProtocolOverrideRecordSet([{
            'protocolKey': 'TestCQM',
            'adjustment': {
                'reference': '2018-02-01T00:00:00Z',
                'cycleDays': 180
            },
            'snooze': None,
            'modified': '2018-03-15T12:32:54Z',
        }])  # yapf: disable
        tested = TestCQM(patient=patient, now=arrow.get('2018-10-30 12:34'))
        result = tested.period_adjustment
        self.assertEqual('2018-02-01T00:00:00Z', result['reference'])
        self.assertEqual(180, result['cycleDays'])

    def test_in_initial_population(self):
        patient = self.load_patient('full')
        # base class
        tested = CQMImplementation(patient=patient)
        result = tested.in_initial_population()
        self.assertIsNone(result)
        # implementation
        tested = TestCQM(patient=patient)
        result = tested.in_initial_population()
        self.assertTrue(result)

    def test_in_denominator(self):
        patient = self.load_patient('full')
        # base class
        tested = CQMImplementation(patient=patient)
        result = tested.in_denominator()
        self.assertIsNone(result)
        # implementation
        tested = TestCQM(patient=patient)
        result = tested.in_denominator()
        self.assertTrue(result)

    def test_in_numerator(self):
        patient = self.load_patient('full')
        # base class
        tested = CQMImplementation(patient=patient)
        result = tested.in_numerator()
        self.assertIsNone(result)
        # implementation
        tested = TestCQM(patient=patient)
        result = tested.in_numerator()
        self.assertTrue(result)

    def test_compute_results(self):
        patient = self.load_patient('full')
        # base class
        tested = CQMImplementation(patient=patient)
        result = tested.compute_results()
        self.assertIsNone(result)
        # implementation
        tested = TestCQM(patient=patient)
        result = tested.compute_results()
        self.assertIsInstance(result, ProtocolResult)
        self.assertEqual('Current', result.status)
        self.assertEqual('This is an example only', result.narrative)
        self.assertEqual(2, len(result.recommendations))
        self.assertIsInstance(result.recommendations[0], TestAResult)
        self.assertIsInstance(result.recommendations[1], TestBResult)
        # change the status --> the computation is independent of the cache
        tested.the_status = 'Pending'
        result = tested.compute_results()
        self.assertIsInstance(result, ProtocolResult)
        self.assertEqual('Pending', result.status)

    def test_results(self):
        patient = self.load_patient('full')
        # base class
        tested = CQMImplementation(patient=patient)
        result = tested.results()
        self.assertIsNone(result)
        # implementation
        tested = TestCQM(patient=patient)
        result = tested.results()
        self.assertIsInstance(result, ProtocolResult)
        self.assertEqual('Current', result.status)
        self.assertEqual('This is an example only', result.narrative)
        self.assertEqual(2, len(result.recommendations))
        self.assertIsInstance(result.recommendations[0], TestAResult)
        self.assertIsInstance(result.recommendations[1], TestBResult)
        # change the status --> the cache is not changed
        tested.the_status = 'Pending'
        result = tested.results()
        self.assertIsInstance(result, ProtocolResult)
        self.assertEqual('Current', result.status)

    def test_recommendations(self):
        patient = self.load_patient('full')
        # base class
        tested = CQMImplementation(patient=patient)
        with self.assertRaises(AttributeError):
            tested.recommendations()
        # implementation
        tested = TestCQM(patient=patient)
        result = tested.recommendations()
        self.assertEqual(2, len(result))
        self.assertIsInstance(result[0], TestAResult)
        self.assertIsInstance(result[1], TestBResult)

    def test_narrative(self):
        patient = self.load_patient('full')
        # base class
        tested = CQMImplementation(patient=patient)
        with self.assertRaises(AttributeError):
            tested.narrative()
        # implementation
        tested = TestCQM(patient=patient)
        result = tested.narrative()
        self.assertEqual('This is an example only', result)

    def test_status(self):
        patient = self.load_patient('full')
        # base class
        tested = CQMImplementation(patient=patient)
        with self.assertRaises(AttributeError):
            tested.status()
        # implementation
        tested = TestCQM(patient=patient)
        result = tested.status()
        self.assertEqual('Current', result)

    def test_due_in(self):
        patient = self.load_patient('full')
        # base class
        tested = CQMImplementation(patient=patient)
        with self.assertRaises(AttributeError):
            tested.due_in()
        # implementation
        tested = TestCQM(patient=patient)
        result = tested.due_in()
        self.assertEqual(-17, result)

    def test_days_of_notice(self):
        patient = self.load_patient('full')
        # base class
        tested = CQMImplementation(patient=patient)
        with self.assertRaises(AttributeError):
            tested.days_of_notice()
        # implementation
        tested = TestCQM(patient=patient)
        result = tested.days_of_notice()
        self.assertEqual(30, result)

    def test_next_review(self):
        patient = self.load_patient('full')
        # base class
        tested = CQMImplementation(patient=patient)
        with self.assertRaises(AttributeError):
            tested.next_review()
        # implementation
        tested = TestCQM(patient=patient)
        result = tested.next_review()
        self.assertEqual('2018-11-17T12:34:56.421', result)

        # - no due_in defined
        tested = TestCQM(patient=patient)
        tested._results = ProtocolResult()
        self.assertIsNone(tested.next_review())

        # - due_in with a negative value
        tested._results.due_in = -1
        self.assertIsNone(tested.next_review())

        # - due_in with a positive value
        tested._results.due_in = 10
        template = '%Y-%m-%d %H:%M:%S'
        expected = (datetime.utcnow() + timedelta(days=10)).strftime(template)
        template = 'YYYY-MM-DD HH:mm:ss'
        result = tested.next_review().format(template)
        self.assertEqual(expected, result)

    def test_display_date(self):
        patient = self.load_patient('full')
        # base class
        tested = CQMImplementation(patient=patient, now=arrow.get('2018-10-01'))

        tests = [
            ('2018-09-09', '3 weeks ago on 9/9/18'),
            ('2018-09-19', 'a week ago on 9/19/18'),
            ('2018-07-13', '3 months ago on 7/13/18'),
            ('2018-03-13', '7 months ago on 3/13/18'),
            ('2017-03-13', 'a year ago on 3/13/17'),
        ]
        for day, expected in tests:
            result = tested.display_date(arrow.get(day))
            self.assertEqual(expected, result)

    def test_display_period(self):
        tests = [
            ('2018-09-09', '2018-12-19', 'between 9/9/18 and 12/19/18$'),
            ('2018-07-17', '2018-08-03', 'between 7/17/18 and 8/3/18$'),
            ('2002-05-21', '2007-02-03', 'between 5/21/02 and 2/3/07$'),
        ]
        for day_from, day_to, expected in tests:
            result = ClinicalQualityMeasure.display_period(arrow.get(day_from), arrow.get(day_to))
            self.assertRegex(result, expected)

    def test_screening_interval_context(self):
        patient = self.load_patient('full')
        # no default_display_interval_in_days set
        result = ClinicalQualityMeasure(patient=patient).screening_interval_context()
        self.assertEqual(result, '')

        class CQMWithDefaultDisplayInterval(ClinicalQualityMeasure):
            class Meta:
                default_display_interval_in_days = 30

        # no period adjustment
        cqm = CQMWithDefaultDisplayInterval(patient=patient)
        result = cqm.screening_interval_context()
        self.assertEqual(result, 'Current screening interval 1 month.')
        # there is a period adjustment
        cqm = CQMWithDefaultDisplayInterval(patient=patient)
        cqm._meta.default_display_interval_in_days = 30
        cqm.period_adjustment = {'cycleDays': 60}
        result = cqm.screening_interval_context()
        self.assertEqual(result, 'Current screening interval 2 months.')

    def test_friendly_time_duration(self):
        tests = [
            (2, '2 days'),
            (30, '1 month'),
            (31, '1 month, 1 day'),
            ((30 * 2), '2 months'),
            (((30 * 2) + 2), '2 months, 2 days'),
            (365, '1 year'),
            (365 + 1, '1 year'),
            (365 + 30, '1 year, 1 month'),
            (365 + 60, '1 year, 2 months'),
            ((365 * 2) + (30 * 2), '2 years, 2 months'),
        ]
        for actual_input, expected_output in tests:
            result = ClinicalQualityMeasure.friendly_time_duration(actual_input)
            self.assertRegex(result, expected_output)


# class to be able to test the top level procedures
class TestProcedures(SDKBaseTest):

    def test_get_subclasses(self):
        result = get_subclasses(TestClassBase)
        self.assertEqual(3, len(result))
        expected = ['TestClassSubA', 'TestClassSubB', 'TestClassSubC']
        for a_class in result:
            self.assertIn(a_class.__name__, expected)

    def test_get_protocols(self):
        # result includes also the actual classes
        result = LOADED_PROTOCOLS
        self.assertEqual(2, len(result))
        self.assertIn('HEALTH_MAINTENANCE', result)
        health = list(map(lambda x: x.__name__, result['HEALTH_MAINTENANCE']))
        self.assertIn('ProtoUnitTestAllTrue', health)
        self.assertNotIn('ProtoUnitTestX', health)
        self.assertIn('ProtoUnitTestY', health)

        expected = [
            'ProtoUnitTestAllTrue',
            'ProtoUnitTestX',
            'ProtoUnitTestY',
            'ProtoUnitTestZ',
        ]
        self.assertIn('UNIT_TESTS', result)
        self.assertEqual(len(expected), len(result['UNIT_TESTS']))
        for a_class in result['UNIT_TESTS']:
            self.assertIn(a_class.__name__, expected)

    def test_protocols_for_patient(self):
        patient = self.load_patient('full')
        protocols = LOADED_PROTOCOLS
        result = protocols_for_patient(protocols['UNIT_TESTS'], patient)
        expected = [
            'ProtoUnitTestAllTrue',
            'ProtoUnitTestX',
            'ProtoUnitTestZ',
        ]
        # ProtoUnitTestY has not_applicable as status
        # ProtoUnitTestZ has not_applicable as status but has due_in not null
        self.assertEqual(len(expected), len(result))

        for instance in result:
            self.assertIn(instance.__class__.__name__, expected)
            self.assertEqual('report', instance.context)

        result = protocols_for_patient(protocols['UNIT_TESTS'], patient, context='guidance')
        expected = [
            'ProtoUnitTestAllTrue',
            'ProtoUnitTestX',
            'ProtoUnitTestZ',
        ]
        # ProtoUnitTestY has not_applicable as status
        # ProtoUnitTestZ has not_applicable as status but has due_in not null
        self.assertEqual(len(expected), len(result))

        for instance in result:
            self.assertIn(instance.__class__.__name__, expected)
            self.assertEqual('guidance', instance.context)

        # no empty change types
        result = protocols_for_patient(protocols['UNIT_TESTS'], patient, [
            'change_A',
            'change_B',
        ])
        expected = [
            'ProtoUnitTestAllTrue',
            'ProtoUnitTestX',
            'ProtoUnitTestZ',
        ]
        # ProtoUnitTestX is impacted by change_C
        # ProtoUnitTestY has not_applicable as status
        # ProtoUnitTestZ has not_applicable as status but has due_in not null
        self.assertEqual(len(expected), len(result))

        for instance in result:
            self.assertIn(instance.__class__.__name__, expected)
            if instance.__class__.__name__ == 'ProtoUnitTestX':
                self.assertEqual('not_relevant', instance.status())
            else:
                self.assertNotEqual('not_relevant', instance.status())

        # ProtoUnitTestAllTrue snoozed
        patient = self.load_patient('full')
        patient.protocol_overrides = ProtocolOverrideRecordSet([{
            'protocolKey': 'ProtoUnitTestAllTrue',
            'adjustment': None,
            'snooze': {
                'reference': arrow.now(),
                'snoozedDays': 22
            },
            'modified': '2018-10-02T21:50:09.661490Z'
        }])

        protocols = LOADED_PROTOCOLS
        result = protocols_for_patient(protocols['UNIT_TESTS'], patient)
        expected = [
            'ProtoUnitTestAllTrue',
            'ProtoUnitTestX',
            'ProtoUnitTestZ',
        ]
        # ProtoUnitTestY has not_applicable as status
        # ProtoUnitTestZ has not_applicable as status but has due_in not null
        self.assertEqual(len(expected), len(result))

        for instance in result:
            class_name = instance.__class__.__name__
            self.assertIn(class_name, expected)
            if class_name == 'ProtoUnitTestAllTrue':
                self.assertTrue(instance.snoozed(), class_name)
            else:
                self.assertFalse(instance.snoozed(), class_name)


# ---helper ---
class TestProto(Protocol):

    __test__ = False

    the_status = 'current status'

    class Meta:

        title = 'Title TestProtocol'

        references = [
            'Small title #1. www.unittests.med/canvas?id=1',
            'Small title #2. http://www.unittests.med/canvas?id=2',
            'Small title #3. https://www.unittests.med/canvas?id=3',
            'Small title #4. WWW.unittests.med/canvas?id=4',
            'Not valid #5  www.unittests.med/canvas?id=5 because of this',
        ]

    def status(self) -> str:
        return self.the_status


class TestPFalse(Protocol):

    __test__ = False

    references = []

    def status(self) -> str:
        return 'Current'

    @classmethod
    def enabled(cls) -> bool:
        return False


class CQMImplementation(ClinicalQualityMeasure):

    class Meta:
        compute_on_change_types = [
            'change_A',
            'change_B',
            'change_C',
        ]

    def in_initial_population(self) -> bool:
        pass

    def in_denominator(self) -> bool:
        pass

    def in_numerator(self) -> bool:
        pass

    def compute_results(self) -> ProtocolResult:
        pass


class TestCQM(ClinicalQualityMeasure):

    __test__ = False

    is_in_ini = True
    is_in_den = True
    is_in_num = True
    the_status = 'Current'

    identifiers = ['TestCQM']

    references = [
        'Small title #1. www.unittests.med/canvas?id=1',
    ]

    @classmethod
    def revision(cls) -> str:
        return '2019-02-12 09:54:31+00:00'

    def in_initial_population(self) -> bool:
        return self.is_in_ini

    def in_denominator(self) -> bool:
        return self.is_in_den

    def in_numerator(self) -> bool:
        return self.is_in_num

    def compute_results(self) -> ProtocolResult:
        result = ProtocolResult()
        result.status = self.the_status
        result.add_narrative('This is an example only')
        result.add_recommendation(TestAResult())
        result.add_recommendation(TestBResult())
        result.due_in = -17
        result.next_review = '2018-11-17T12:34:56.421'
        return result


class TestAResult(Recommendation):

    __test__ = False

    def __init__(self):
        command = {'type': 'TestAResult'}

        super().__init__(
            key='KEY-ID',
            rank=123,
            button='ACT',
            title='Title TestAResult',
            narrative='Narrative TestAResult',
            command=command)


class TestBResult(Recommendation):

    __test__ = False

    def __init__(self):
        command = {'type': 'TestBResult'}

        super().__init__(
            key='KEY-ID',
            rank=123,
            button='ACT',
            title='Title TestBResult',
            narrative='Narrative TestBResult',
            command=command)


class TestClassBase():

    __test__ = False


class TestClassSubA(TestClassBase):

    __test__ = False


class TestClassSubB(TestClassBase):

    __test__ = False


class TestClassSubC(TestClassSubB):

    __test__ = False



import arrow


class Timeframe(object):

    def __init__(self, start: arrow.Arrow, end: arrow.Arrow):
        self.start = start
        self.end = end

    def __str__(self):
        return f'<Timeframe start={self.start}, end={self.end}>'

    # the duration of the time frame defines the duration
    @property
    def duration(self) -> int:
        return (self.end - self.start).days

    def increased_by(self, years: int = 0, months: int = 0, days: int = 0):
        start = self.start
        end = self.end

        if years > 0:
            end = end.shift(years=years)
        elif years < 0:
            start = start.shift(years=years)

        if months > 0:
            end = end.shift(months=months)
        elif months < 0:
            start = start.shift(months=months)

        if days > 0:
            end = end.shift(days=days)
        elif days < 0:
            start = start.shift(days=days)

        return Timeframe(start=start, end=end)



from unittest import TestCase

from canvas_workflow_kit.builtin_cqms.helper_population import Population


class TestPopulation(TestCase):

    def test___init__(self):
        tested = Population()
        self.assertIsNone(tested.in_initial_population)
        self.assertIsNone(tested.in_denominator)
        self.assertIsNone(tested.in_numerator)

    def test_set_initial_population(self):
        tested = Population()

        tested.set_initial_population(True)
        self.assertTrue(tested.in_initial_population)
        self.assertTrue(tested.in_denominator)
        self.assertTrue(tested.in_numerator)

        tested.set_initial_population(False)
        self.assertFalse(tested.in_initial_population)
        self.assertFalse(tested.in_denominator)
        self.assertFalse(tested.in_numerator)

    def test_set_denominator(self):
        tested = Population()

        tested.set_denominator(True)
        self.assertIsNone(tested.in_initial_population)
        self.assertTrue(tested.in_denominator)
        self.assertTrue(tested.in_numerator)

        tested.set_denominator(False)
        self.assertIsNone(tested.in_initial_population)
        self.assertFalse(tested.in_denominator)
        self.assertFalse(tested.in_numerator)

    def test_set_numerator(self):
        tested = Population()

        tested.set_numerator(True)
        self.assertIsNone(tested.in_initial_population)
        self.assertIsNone(tested.in_denominator)
        self.assertTrue(tested.in_numerator)

        tested.set_numerator(False)
        self.assertIsNone(tested.in_initial_population)
        self.assertIsNone(tested.in_denominator)
        self.assertFalse(tested.in_numerator)



from unittest import TestCase
from canvas_workflow_kit.value_set import v2018


class SDKValueSetTest(TestCase):

    def test_basics(self):
        self.assertTrue(v2018.Diabetes)
        self.assertIn('E1010', v2018.Diabetes.values['icd10cm'])

    def test_or(self):
        diabetes_or_cough = v2018.UpperRespiratoryInfection | v2018.Diabetes
        self.assertIn('E1010', diabetes_or_cough.values['icd10cm'])
        self.assertIn('195708003', diabetes_or_cough.values['snomedct'])

        diabetes_or_cough_or_more = v2018.UpperRespiratoryInfection | v2018.Diabetes | v2018.IschemicVascularDisease
        self.assertIn('E1010', diabetes_or_cough_or_more.values['icd10cm'])
        self.assertIn('195708003', diabetes_or_cough_or_more.values['snomedct'])
        self.assertIn('I70261', diabetes_or_cough_or_more.values['icd10cm'])



import arrow

from canvas_workflow_kit import builtin_cqms
from canvas_workflow_kit.builtin_cqms.cms122v6_diabetes_hemoglobin_a1c_poor_control import (
    ClinicalQualityMeasure122v6
)
from canvas_workflow_kit.builtin_cqms.cms123v6_diabetes_foot_exam import ClinicalQualityMeasure123v6
from canvas_workflow_kit.builtin_cqms.cms131v6_diabetes_eye_exam import ClinicalQualityMeasure131v6
from canvas_workflow_kit.builtin_cqms.cms134v6_diabetes_medical_attention_for_nephropathy import (
    ClinicalQualityMeasure134v6
)
from canvas_workflow_kit.protocol import get_protocols  # , STATUS_NOT_APPLICABLE
from canvas_workflow_kit.recommendation import (
    InterviewRecommendation,
    LabRecommendation,
    PerformRecommendation,
    ReferRecommendation
)
from canvas_workflow_kit.timeframe import Timeframe

from .base import SDKBaseTest

# from protocols.value_set import v2018


class SDKProtocolTest(SDKBaseTest):

    def setUp(self):
        super().setUp()
        self.protocols = get_protocols(builtin_cqms)

    def test_cms122v6(self):
        now = arrow.get('2018-08-23 13:24:56')
        # diabetic1 hasn't had a hab1c test, should tell him to get one
        patient = self.load_patient('diabetic1')
        protocol = ClinicalQualityMeasure122v6(now=now, patient=patient)
        self.assertTrue(protocol.in_denominator())

        recs = protocol.recommendations()
        self.assertEqual(len(recs), 1)
        self.assertTrue(isinstance(recs[0], LabRecommendation))
        self.assertEqual(protocol.status(), 'due')
        self.assertIn('last hba1c test', protocol.narrative().lower())

        # diabetic2 has had a test but it was too long ago, same result
        patient = self.load_patient('diabetic2')
        protocol = ClinicalQualityMeasure122v6(now=now, patient=patient)
        self.assertTrue(protocol.in_denominator())

        recs = protocol.recommendations()
        self.assertEqual(len(recs), 1)
        self.assertTrue(isinstance(recs[0], LabRecommendation))
        self.assertEqual(protocol.status(), 'due')
        self.assertIn('last hba1c test', protocol.narrative().lower())

        # diabetic3 is doing great, good job diabetic3!
        patient = self.load_patient('diabetic3')
        protocol = ClinicalQualityMeasure122v6(now=now, patient=patient)
        self.assertTrue(protocol.in_denominator())

        recs = protocol.recommendations()
        self.assertEqual(len(recs), 0)
        self.assertEqual(protocol.status(), 'satisfied')
        self.assertIn('last hba1c', protocol.narrative().lower())

        # diabetic4 had a recent test but it was bad.
        patient = self.load_patient('diabetic4')
        protocol = ClinicalQualityMeasure122v6(now=now, patient=patient)
        self.assertTrue(protocol.in_denominator())

        recs = protocol.recommendations()
        self.assertEqual(len(recs), 1)
        self.assertEqual(protocol.status(), 'due')
        self.assertIn('last hba1c', protocol.narrative().lower())

    def test_cms123v6(self):
        # not yet screened, should get a screening
        patient = self.load_patient('diabetic1')
        now = arrow.get('2017-11-01 12:00:00')
        timeframe = Timeframe(
            start=now.replace(hour=0, minute=0), end=now.replace(hour=23, minute=59))
        protocol = ClinicalQualityMeasure123v6(patient=patient, timeframe=timeframe)
        self.assertTrue(protocol.in_denominator())

        recs = protocol.recommendations()
        self.assertEqual(len(recs), 1)
        self.assertTrue(isinstance(recs[0], InterviewRecommendation))
        # self.assertTrue(isinstance(recs[1], InterviewRecommendation))
        # self.assertTrue(isinstance(recs[2], InterviewRecommendation))
        self.assertEqual(protocol.status(), 'due')
        self.assertIn('foot', protocol.narrative().lower())

        # got screened, satisfied
        patient = self.load_patient('feet1')
        protocol = ClinicalQualityMeasure123v6(patient=patient, timeframe=timeframe)
        self.assertTrue(protocol.in_denominator())

        recs = protocol.recommendations()
        self.assertEqual(len(recs), 0)
        self.assertEqual(protocol.status(), 'satisfied')

    def test_cms134v6(self):
        # not yet tested, should get a test
        patient = self.load_patient('diabetic1')
        now = arrow.get('2017-11-01 12:00:00')
        timeframe = Timeframe(
            start=now.replace(hour=0, minute=0), end=now.replace(hour=23, minute=59))
        protocol = ClinicalQualityMeasure134v6(patient=patient, timeframe=timeframe)
        self.assertTrue(protocol.in_denominator())

        recs = protocol.recommendations()
        self.assertEqual(len(recs), 1)
        self.assertTrue(isinstance(recs[0], LabRecommendation))
        self.assertEqual(protocol.status(), 'due')
        self.assertIn('urine', protocol.narrative().lower())

    def test_cms131v6(self):
        # not yet tested, should get a test
        patient = self.load_patient('diabetic1')
        now = arrow.get('2017-11-01 12:00:00')
        timeframe = Timeframe(
            start=now.replace(hour=0, minute=0), end=now.replace(hour=23, minute=59))
        protocol = ClinicalQualityMeasure131v6(patient=patient, timeframe=timeframe)
        self.assertTrue(protocol.in_denominator())

        recs = protocol.recommendations()
        self.assertEqual(len(recs), 2)
        self.assertTrue(isinstance(recs[0], PerformRecommendation))
        self.assertTrue(isinstance(recs[1], ReferRecommendation))
        self.assertEqual(protocol.status(), 'due')
        self.assertIn('retinal', protocol.narrative().lower())



# type: ignore
from typing import Type

from cached_property import cached_property

from canvas_workflow_kit import events
from canvas_workflow_kit.protocol import (
    STATUS_DUE,
    STATUS_SATISFIED,
    ClinicalQualityMeasure,
    ProtocolResult
)
from canvas_workflow_kit.recommendation import Recommendation
from canvas_workflow_kit.value_set.specials import (
    DiabetesCirculatoryClassConditionSuspect,
    DiabetesEyeClassConditionSuspect,
    DiabetesEyeConditionSuspect,
    DiabetesNeurologicConditionSuspect,
    DiabetesOtherClassConditionSuspect,
    DiabetesRenalConditionSuspect,
    DiabetesWithoutComplication
)
from canvas_workflow_kit.value_set.value_set import ValueSet


class Hcc003v1(ClinicalQualityMeasure):

    class Meta:
        title = 'Diabetes Mellitus With Secondary Complication Suspect'
        version = '2019-02-12v1'
        description = 'All patients with diabetes, uncomplicated AND a 2ndary condition often associated with diabetes.'  # noqa: E501
        information = 'https://canvas-medical.zendesk.com/hc/en-us/articles/360057221174-Diabetes-Mellitus-With-Secondary-Complication-Suspect-HCC003v1'  # noqa: E501

        identifiers = ['HCC003v1']

        types = ['HCC']

        responds_to_event_types = [
            events.HEALTH_MAINTENANCE,
        ]
        authors = [
            'Canvas Medical Team',
        ]

        references = [
            'Canvas Medical HCC, https://canvas-medical.zendesk.com/hc/en-us/articles/360057221174-Diabetes-Mellitus-With-Secondary-Complication-Suspect-HCC003v1'  # noqa: E501
        ]

        compute_on_change_types = [
            ClinicalQualityMeasure.CHANGE_PROTOCOL_OVERRIDE,
            ClinicalQualityMeasure.CHANGE_CONDITION,
        ]

    def _has_active_condition(self, conditions: Type[ValueSet]) -> bool:
        if self.patient.conditions.intersects(
                self.timeframe, still_active=self.patient.active_only).find(conditions):
            return True
        return False

    def _has_active_class_condition(self, class_conditions: Type[ValueSet]) -> bool:
        if (self.patient.conditions
                        .intersects(self.timeframe, still_active=self.patient.active_only)
                        .find_class(class_conditions)):  # yapf: disable
            return True
        return False

    @property
    def has_diabetes_with_unspecified_condition(self) -> bool:
        return self._has_active_condition(DiabetesWithoutComplication)

    @cached_property
    def has_suspect_eye_condition(self) -> bool:
        return (self._has_active_condition(DiabetesEyeConditionSuspect) |
                self._has_active_class_condition(DiabetesEyeClassConditionSuspect))

    @cached_property
    def has_suspect_neurologic_condition(self) -> bool:
        return self._has_active_condition(DiabetesNeurologicConditionSuspect)

    @cached_property
    def has_suspect_renal_condition(self) -> bool:
        return self._has_active_condition(DiabetesRenalConditionSuspect)

    @cached_property
    def has_suspect_circulatory_condition(self) -> bool:
        return self._has_active_class_condition(DiabetesCirculatoryClassConditionSuspect)

    @cached_property
    def has_suspect_other_condition(self) -> bool:
        return self._has_active_class_condition(DiabetesOtherClassConditionSuspect)

    def in_initial_population(self) -> bool:
        return True

    def in_denominator(self) -> bool:
        """
        Patients with active condition E11.9 Diabetes 2 with unspecified complications
        """
        return self.has_diabetes_with_unspecified_condition

    def in_numerator(self) -> bool:
        """
        Patients with active Conditions as potential Diabetes complication
        """
        return (self.has_suspect_eye_condition or self.has_suspect_neurologic_condition or
                self.has_suspect_renal_condition or self.has_suspect_circulatory_condition or
                self.has_suspect_other_condition)

    def compute_results(self) -> ProtocolResult:
        result = ProtocolResult()

        if self.in_denominator():
            if self.in_numerator():
                result.due_in = -1
                result.status = STATUS_DUE

                if self.has_suspect_eye_condition:
                    comment = (
                        '{0} has Diabetes without complications AND '
                        'an eye condition commonly caused by diabetes on the Conditions list.'
                    ).format(self.patient.first_name)
                    title = ('Consider updating the Diabetes without complications (E11.9) '
                             'to Diabetes with secondary eye disease as clinically appropriate.')
                    result.add_narrative(comment)
                    result.add_recommendation(
                        Recommendation(
                            key='HCC003v1_RECOMMEND_DIAGNOSE_EYE',
                            rank=1,
                            button='Diagnose',
                            title=title,
                            narrative=comment,
                            command={'key': 'diagnose'}))

                if self.has_suspect_neurologic_condition:
                    comment = (
                        '{0} has Diabetes without complications AND '
                        'a neurological condition commonly caused by diabetes on the Conditions list.'  # noqa: E501
                    ).format(self.patient.first_name)
                    title = (
                        'Consider updating the Diabetes without complications (E11.9) '
                        'to Diabetes with secondary neurological sequela as clinically appropriate.'
                    )
                    result.add_narrative(comment)
                    result.add_recommendation(
                        Recommendation(
                            key='HCC003v1_RECOMMEND_DIAGNOSE_NEUROLOGICAL_SEQUELA',
                            rank=2,
                            button='Diagnose',
                            title=title,
                            narrative=comment,
                            command={'key': 'diagnose'}))

                if self.has_suspect_renal_condition:
                    comment = (
                        '{0} has Diabetes without complications AND '
                        'a chronic renal condition commonly caused by diabetes on the Conditions list.'  # noqa: E501
                    ).format(self.patient.first_name)
                    title = ('Consider updating the Diabetes without complications (E11.9) '
                             'to Diabetes with secondary renal disease as clinically appropriate.')
                    result.add_narrative(comment)
                    result.add_recommendation(
                        Recommendation(
                            key='HCC003v1_RECOMMEND_DIAGNOSE_RENAL_DISEASE',
                            rank=3,
                            button='Diagnose',
                            title=title,
                            narrative=comment,
                            command={'key': 'diagnose'}))

                if self.has_suspect_circulatory_condition:
                    comment = (
                        '{0} has Diabetes without complications AND '
                        'a circulatory condition commonly caused by diabetes on the Conditions list.'  # noqa: E501
                    ).format(self.patient.first_name)
                    title = (
                        'Consider updating the Diabetes without complications (E11.9) '
                        'to Diabetes with secondary circulatory disorder as clinically appropriate.'  # noqa: E501
                    )
                    result.add_narrative(comment)
                    result.add_recommendation(
                        Recommendation(
                            key='HCC003v1_RECOMMEND_DIAGNOSE_CIRCULATORY_DISORDER',
                            rank=4,
                            button='Diagnose',
                            title=title,
                            narrative=comment,
                            command={'key': 'diagnose'}))

                if self.has_suspect_other_condition:
                    comment = (
                        '{0} has Diabetes without complications AND '
                        'an another condition commonly caused by diabetes on the Conditions list.'
                    ).format(self.patient.first_name)
                    title = (
                        'Consider updating the Diabetes without complications (E11.9) '
                        'to Diabetes with other secondary complication as clinically appropriate.')
                    result.add_narrative(comment)
                    result.add_recommendation(
                        Recommendation(
                            key='HCC003v1_RECOMMEND_DIAGNOSE_COMPLICATION',
                            rank=5,
                            button='Diagnose',
                            title=title,
                            narrative=comment,
                            command={'key': 'diagnose'}))

            else:
                result.status = STATUS_SATISFIED
        return result



# type: ignore
from typing import List

import arrow

from canvas_workflow_kit import events
from canvas_workflow_kit.protocol import (
    STATUS_DUE,
    STATUS_SATISFIED,
    ClinicalQualityMeasure,
    ProtocolResult
)
from canvas_workflow_kit.recommendation import LabRecommendation
from canvas_workflow_kit.value_set.medication_class_path2018 import AceInhibitors
from canvas_workflow_kit.value_set.specials import CMS134v6Dialysis
# flake8: noqa
from canvas_workflow_kit.value_set.v2018 import (
    Diabetes,
    DiabeticNephropathy,
    DialysisEducation,
    GlomerulonephritisAndNephroticSyndrome,
    HypertensiveChronicKidneyDisease,
    KidneyFailure,
    KidneyTransplant,
    Proteinuria,
    UrineProteinTests
)
from .diabetes_quality_measure import DiabetesQualityMeasure


class ClinicalQualityMeasure134v6(DiabetesQualityMeasure):
    """
    Diabetes: Medical Attention for Nephropathy

    Description: The percentage of patients 18-75 years of age with diabetes who had a nephropathy
    screening test or evidence of nephropathy during the measurement period

    Definition: None

    Rationale: As the seventh leading cause of death in the U.S., diabetes kills approximately
    75,000 people a year (CDC FastStats 2015). Diabetes is a group of diseases marked by high blood
    glucose levels, resulting from the body's inability to produce or use insulin (CDC Statistics
    2014, ADA Basics 2013). People with diabetes are at increased risk of serious health
    complications including vision loss, heart disease, stroke, kidney failure, amputation of toes,
    feet or legs, and premature death. (CDC Fact Sheet 2014).

    In 2012, diabetes cost the U.S. an estimated $245 billion: $176 billion in direct medical costs
    and $69 billion in reduced productivity. This is a 41 percent increase from the estimated $174
    billion spent on diabetes in 2007 (ADA Economic 2013).

    In 2011, diabetes accounted for 44% of new kidney failure cases. In the same year, 49,677
    diabetics started treatment for kidney failure and 228,924 people of all ages with kidney
    failure due to diabetes were living on chronic dialysis or with a kidney transplant (CDC
    Statistics, 2014).

    Guidance: Only patients with a diagnosis of Type 1 or Type 2 diabetes should be included in the
    denominator of this measure; patients with a diagnosis of secondary diabetes due to another
    condition should not be included

    More information: https://ecqi.healthit.gov/sites/default/files/ecqm/measures/CMS134v6.html
    """

    class Meta:
        title = 'Diabetes: Medical Attention for Nephropathy'
        version = '2019-02-12v1'
        description = (
            'Patients 18-75 years of age with diabetes who have not had a nephropathy screening test '
            'in the last year or evidence of nephropathy.')
        information = 'https://ecqi.healthit.gov/sites/default/files/ecqm/measures/CMS134v6.html'

        identifiers = ['CMS134v6']

        types = ['CQM']

        responds_to_event_types = [
            events.HEALTH_MAINTENANCE,
        ]

        authors = [
            'National Committee for Quality Assurance',
        ]

        references = [
            'American Diabetes Association. Microvascular complications and foot care. Sec. 10. In Standards of Medical Care in Diabetes-2017. Diabetes Care 2017;40(Suppl. 1):S88-S98.',
            'American Diabetes Association. 2013. Diabetes Basics. www.diabetes.org/diabetes-basics/?loc=GlobalNavDB.',
            'American Diabetes Association (ADA). April 2013. Economic Costs of Diabetes in the U.S. in 2012. Diabetes Care. Vol. 36 no. 4 1033-46. http://care.diabetesjournals.org/content/36/4/1033.full.',
            'Centers for Disease Control and Prevention (CDC). 2014. National Diabetes Statistics Report. http://www.cdc.gov/diabetes/pdfs/data/2014-report-estimates-of-diabetes-and-its-burden-in-the-united-states.pdf',
            'Centers for Disease Control and Prevention (CDC). 2015. FastStats: Deaths and Mortality. www.cdc.gov/nchs/fastats/deaths.htm.',
            'Centers for Disease Control and Prevention. 2014. CDC Features. Diabetes Latest. www.cdc.gov/features/diabetesfactsheet/.',
            'Handelsman Y, Bloomgarden ZT, Grunberger G, Umpierrez G, Zimmerman RS, Bailey TS, et al. (2015) American Association of Clinical Endocrinologists and American College of Endocrinology-Clinical Practice Guidelines for Developing a Diabetes Mellitus Comprehensive Care Plan-2015. Endocr Pract 21 Suppl 1: 1-87.',
        ]

        compute_on_change_types = [
            ClinicalQualityMeasure.CHANGE_PROTOCOL_OVERRIDE,
            ClinicalQualityMeasure.CHANGE_CONDITION,
            ClinicalQualityMeasure.CHANGE_INSTRUCTION,
            ClinicalQualityMeasure.CHANGE_LAB_REPORT,
            ClinicalQualityMeasure.CHANGE_MEDICATION,
            ClinicalQualityMeasure.CHANGE_PATIENT,
            ClinicalQualityMeasure.CHANGE_REFERRAL_REPORT,
        ]

    @classmethod
    def enabled(cls) -> bool:
        return True

    message: str = None
    _due_in: int = -1

    DISMISSING_CONDITIONS = [
        (HypertensiveChronicKidneyDisease, 'Hypertensive Chronic Kidney Disease'),
        (KidneyFailure, 'Kidney Failure'),
        (GlomerulonephritisAndNephroticSyndrome, 'Glomerulonephritis and Nephrotic Syndrome'),
        (DiabeticNephropathy, 'Diabetic Nephropathy'),
        (Proteinuria, 'Proteinuria'),
    ]

    def in_denominator(self) -> bool:
        """
        Denominator: Equals Initial Population

        Exclusions: Exclude patients who were in hospice care during the measurement year

        Exceptions: None
        """
        if not self.in_initial_population():
            return False

        if self.patient.hospice_within(self.timeframe):
            return False

        return True

    def in_numerator(self) -> bool:
        """
        Numerator: Patients with a screening for nephropathy or evidence of nephropathy during the
        measurement period

        Exclusions: Not Applicable
        """
        self.message = None
        self._due_in = -1

        # VascularAccessForDialysis, DialysisServices and OtherServicesRelatedToDialysis
        #  replaced with CMS134v6Dialysis
        record = self.patient.referral_reports.find(CMS134v6Dialysis).within(self.timeframe).last()
        if record:
            self.message = '{name} has diabetes and had a Dialysis Related Service {date}'.format(
                name=self.patient.first_name,
                date=self.display_date(arrow.get(record['originalDate'])))
            return True

        # medication
        records = (self.patient.medications.find(AceInhibitors).intersects(
            self.timeframe, still_active=self.patient.active_only))
        if records:
            self.message = '{name} has diabetes and is under Ace Inhibitors medication'.format(
                name=self.patient.first_name)
            return True

        # conditions
        for (condition, label) in self.DISMISSING_CONDITIONS:
            records = (self.patient.conditions.find(condition).intersects(
                self.timeframe, still_active=self.patient.active_only))
            if records:
                self.message = '{name} has diabetes and has been diagnosed {label}'.format(
                    name=self.patient.first_name, label=label)
                return True

        records = (self.patient.conditions.find(KidneyTransplant).intersects(
            self.timeframe, still_active=self.patient.active_only))
        if records:
            self.message = '{name} has diabetes and had a Kidney Transplant'.format(
                name=self.patient.first_name)
            return True

        # instruction
        record = (self.patient.instructions.find(DialysisEducation).within(self.timeframe).last())
        if record:
            self.message = '{name} has diabetes and had an ESRD Monthly Outpatient Services {date}'.format(
                name=self.patient.first_name, date=self.display_date(arrow.get(record['noteTimestamp'])))
            return True

        # lab test
        record = self.patient.lab_reports.find(UrineProteinTests).within(self.timeframe).last()
        if record:
            on_date = arrow.get(record['originalDate'])
            self._due_in = (on_date.shift(days=self.timeframe.duration) - self.now).days
            self.message = '{name} has diabetes and a urine protein test was done {date}'.format(
                name=self.patient.first_name, date=self.display_date(on_date))
            return True

        return False

    def compute_results(self) -> ProtocolResult:
        """
        Clinical recommendation: American Diabetes Association (2017):

        Screening
        - At least once a year, quantitatively assess urinary albumin (eg, spot urinary albumin-to-
        creatinine ratio [UACR]) and estimated glomerular filtration rate (eGFR) in patients with
        type 1 diabetes duration of greater than or equal to 5 years in all patients with type 2
        diabetes, and in all patients with comorbid hypertension. (Level of evidence: B)

        Treatment
        - An angiotensin-converting enzyme (ACE) inhibitor or angiotensin receptor blocker (ARB) is
        not recommended for the primary prevention of diabetic kidney disease in patients with
        diabetes who have normal blood pressure, normal UACR (<30 mg/g), and normal estimated
        glomerular filtration rate. (Level of evidence: B)
        - Either an ACE inhibitor or ARB is suggested for the treatment of the nonpregnant patient
        with modestly elevated UACR (30-299 mg/day) (Level of evidence: C) and is strongly
        recommended for those with urinary albumin excretion >=300 mg/day. (Level of evidence: A)
        - When ACE inhibitors, ARBs, or diuretics are used, monitor serum creatinine and potassium
        levels for the development of increased creatinine or changes in potassium. (Level of
        evidence: E)
        - Continued monitoring of UACR in patients with albuminuria treated with an ACE inhibitor
        or ARBs is reasonable to assess progression of diabetic kidney disease. (Level of evidence:
        E)

        American Association of Clinical Endocrinologists (2015):
        - Beginning 5 years after diagnosis in patients with type 1 diabetes (if diagnosed before
        age 30) or at diagnosis in patients with type 2 diabetes and those with type 1 diabetes
        diagnosed after age 30, annual assessment of serum creatinine to determine the estimated
        glomerular filtration rate (eGFR) and urine albumin excretion rate (AER) should be
        performed to identify, stage, and monitor progression of diabetic nephropathy (Grade C;
        best evidence level 3).
        - Patients with nephropathy should be counseled regarding the need for optimal glycemic
        control, blood pressure control, dyslipidemia control, and smoking cessation (Grade B; best
        evidence level 2).
        - In addition, they should have routine monitoring of albuminuria, kidney function
        electrolytes, and lipids (Grade B; best evidence level 2).
        - Associated conditions such as anemia and bone and mineral disorders should be assessed as
        kidney function declines (Grade D; best evidence level 4).
        - Referral to a nephrologist is recommended well before the need for renal replacement
        therapy (Grade D; best evidence level 4).
        """
        result = ProtocolResult()

        if self.in_denominator():
            if self.in_numerator():
                result.status = STATUS_SATISFIED
                result.add_narrative(self.message)
                result.due_in = self._due_in
            else:
                result.due_in = -1
                result.status = STATUS_DUE
                result.add_narrative(
                    '{name} has diabetes and a urine microalbumin test'
                    ' is due to screen for nephropathy'.format(name=self.patient.first_name))
                result.add_recommendation(
                    LabRecommendation(
                        key='CMS134v6_RECOMMEND_URINE_TEST',
                        rank=1,
                        button='Order',
                        patient=self.patient,
                        condition=Diabetes,
                        lab=UrineProteinTests,
                        title='Order a urine microalbumin test'))
        else:
            result.due_in = self.first_due_in()

        return result



import re


def camelcase(string):
    """ Convert string into camel case.
    Args:
        string: String to convert.
    Returns:
        string: Camel case string.
    """

    string = re.sub(r"^[\-_\.]", '', str(lowercase(string)))
    if not string:
        return string
    return lowercase(string[0]) + re.sub(r"[\-_\.\s]([a-z])", lambda matched: uppercase(matched.group(1)), string[1:])


def snakecase(string):
    """Convert string into snake case.
    Join punctuation with underscore
    Args:
        string: String to convert.
    Returns:
        string: Snake cased string.
    """

    string = re.sub(r"[\-\.\s]", '_', str(string))
    if not string:
        return string
    return lowercase(string[0]) + re.sub(r"[A-Z]", lambda matched: '_' + lowercase(matched.group(0)), string[1:])


def uppercase(string):
    """Convert string into upper case.
    Args:
        string: String to convert.
    Returns:
        string: Uppercase case string.
    """

    return str(string).upper()


def lowercase(string):
    """Convert string into lower case.
    Args:
        string: String to convert.
    Returns:
        string: Lowercase case string.
    """

    return str(string).lower()



# type: ignore
from typing import Dict, Optional, cast

import arrow

from cached_property import cached_property

from canvas_workflow_kit import events, settings
from canvas_workflow_kit.protocol import (
    STATUS_DUE,
    STATUS_SATISFIED,
    ClinicalQualityMeasure,
    ExternallyAwareClinicalQualityMeasure,
    ProtocolResult
)
from canvas_workflow_kit.recommendation import StructuredAssessmentRecommendation
from canvas_workflow_kit.timeframe import Timeframe
from canvas_workflow_kit.value_set.customer_value_sets.scan import (
    StructuredAssessmentQuestionnaireMoodDisorder
)
from canvas_workflow_kit.value_set.value_set import ValueSet


class QuestionnairePhq9(ValueSet):

    VALUE_SET_NAME = 'PHQ-9 Questionnaire'

    LOINC = {'44249-1'}


class QuestionnaireGad7(ValueSet):

    VALUE_SET_NAME = 'GAD-7 Questionnaire'

    LOINC = {'69737-5'}


class CustomerScanMoodDisorder(ExternallyAwareClinicalQualityMeasure, ClinicalQualityMeasure):
    """
    Major Depression
    phq9 score of 5 or greater
    Mood Disorder SA
    Please consider the need to include major depression as a diagnosis within the care plan.
    Please use the Mood Disorder Structured Assessment.

    Generalized Anxiety Disorder
    GAD7 score of 5 or greater
    Mood Disorder SA
    Please consider the need to include generalized anxiety disorder as a diagnosis within the care
    plan.  Please use the Mood Disorder Structured Assessment.
    """

    class Meta:
        title = 'SCAN: Mood Disorder Structured Assessment'
        version = '2020-09-04v1'
        description = ''
        information = 'https://docs.google.com/document/d/1NW4b3OT_yPeZy_oIKt_DCCeQ0egDaWFE/edit'

        identifiers = ['CustomerScanMoodDisorderV1']

        types = ['CQM']

        responds_to_event_types = [
            events.HEALTH_MAINTENANCE,
        ]

        authors = [
            'SCAN',
        ]

        compute_on_change_types = [
            ClinicalQualityMeasure.CHANGE_PROTOCOL_OVERRIDE,
            ClinicalQualityMeasure.CHANGE_INTERVIEW,
        ]

        funding_source = 'SCAN'

        references = ['Written by SCAN']

    @classmethod
    def enabled(cls) -> bool:
        return cast(bool, settings.IS_HEALTHCHEC)

    def in_initial_population(self) -> bool:
        return True

    def interview_lookback_timeframe(self) -> Timeframe:
        return Timeframe(start=self.timeframe.end.shift(days=-30), end=self.timeframe.end)

    def last_interview(self, questionnaire_valueset, timeframe=None) -> Optional[Dict]:
        interview = self.patient.interviews.find(questionnaire_valueset).within(
            timeframe or self.timeframe).last()

        if not interview:
            return None

        return interview

    @cached_property
    def last_gad7(self) -> Optional[Dict]:
        return self.last_interview(
            questionnaire_valueset=QuestionnaireGad7,
            timeframe=self.interview_lookback_timeframe())

    def last_gad7_score(self) -> Optional[int]:
        if self.last_gad7:
            return self.last_gad7['results'][0]['score']
        return None

    def last_gad7_date(self) -> Optional[arrow.Arrow]:
        if self.last_gad7:
            return arrow.get(self.last_gad7['noteTimestamp'])
        return None

    @cached_property
    def last_phq9(self) -> Optional[Dict]:
        return self.last_interview(
            questionnaire_valueset=QuestionnairePhq9,
            timeframe=self.interview_lookback_timeframe())

    def last_phq9_score(self) -> Optional[int]:
        if self.last_phq9:
            return self.last_phq9['results'][0]['score']
        return None

    def last_phq9_date(self) -> Optional[arrow.Arrow]:
        if self.last_phq9:
            return arrow.get(self.last_phq9['noteTimestamp'])
        return None

    def last_elevated_gad7_or_phq9_date(self) -> Optional[arrow.Arrow]:
        date_of_most_recent_elevated_screening = None

        if self.last_gad7_score() and self.last_gad7_score() >= 5:
            date_of_most_recent_elevated_screening = self.last_gad7_date()

        if self.last_phq9_score() and self.last_phq9_score() >= 5:
            if date_of_most_recent_elevated_screening and self.last_phq9_date():
                date_of_most_recent_elevated_screening = max(
                    date_of_most_recent_elevated_screening, self.last_phq9_date())
            else:
                date_of_most_recent_elevated_screening = self.last_phq9_date()

        return date_of_most_recent_elevated_screening

    @cached_property
    def last_mood_disorder_assessment(self) -> Optional[Dict]:
        return self.last_interview(
            questionnaire_valueset=StructuredAssessmentQuestionnaireMoodDisorder,
            timeframe=self.interview_lookback_timeframe())

    def last_mood_disorder_assessment_date(self) -> Optional[arrow.Arrow]:
        if self.last_mood_disorder_assessment:
            return arrow.get(self.last_mood_disorder_assessment['noteTimestamp'])
        return None

    def mood_disorder_assessed_on_or_after(self, reference_date: arrow.Arrow) -> bool:
        # It can't have been assessed after a date that doesn't exist!
        if not reference_date:
            return False

        # If it hasn't been assessed, it certainly wasn't assessed after the
        # reference date.
        if not self.last_mood_disorder_assessment_date():
            return False

        return self.last_mood_disorder_assessment_date() >= reference_date

    def in_denominator(self) -> bool:
        """
        Denominator: Equals Initial Population

        Exclusions: Exclude patients who were in hospice care during the measurement year

        Exceptions: None
        """
        if not self.in_initial_population():
            return False

        if self.patient.hospice_within(self.timeframe):
            return False

        return True

    def in_numerator(self) -> bool:
        """
        Patient is in the numerator if their last gad7 or phq9 test was not elevated, or if it was,
        but they have had a mood disorder assessment since then. They are also in the numerator if
        they have not had any gad7 or phq9 result. (Can't be elevated if never assessed!)
        """
        mood_disorder_indicated = False

        date_of_most_recent_elevated_screening = self.last_elevated_gad7_or_phq9_date()

        if date_of_most_recent_elevated_screening:
            mood_disorder_indicated = True

        # Patient is in the numerator if the scores do not indicate a mood
        # disorder. Being in the numerator means the protocol is satisfied.
        # The goal is to ensure patients are monitored appropriately. Getting
        # the numerator as close to the denominator as possible indicates good
        # monitoring.
        mood_disorder_not_indicated = not mood_disorder_indicated

        mood_disorder_assessed = self.mood_disorder_assessed_on_or_after(
            date_of_most_recent_elevated_screening)

        return (mood_disorder_not_indicated or
                (mood_disorder_indicated and mood_disorder_assessed))

    def craft_satisfied_result(self):
        """
        Satisfied if in numerator. (See numerator description)
        """
        result = ProtocolResult()

        result.due_in = -1
        result.status = STATUS_SATISFIED

        mood_disorder_assessed = self.mood_disorder_assessed_on_or_after(
            self.last_elevated_gad7_or_phq9_date())
        if mood_disorder_assessed:
            result.add_narrative(f'{self.patient.first_name} had elevated PHQ9 or GAD7 scores, '
                                 'but a mood disorder assessment was performed '
                                 f'{self.display_date(self.last_elevated_gad7_or_phq9_date())}')
        else:
            result.add_narrative(f'{self.patient.first_name} has not had elevated PHQ9 or GAD7 '
                                 'scores in the past 30 days.')

        return result

    def craft_unsatisfied_result(self):
        """
        Unsatisfied if not in the numerator. Recommend a Mood Disorder Structured Assessment.
        """
        result = ProtocolResult()

        result.due_in = -1
        result.status = STATUS_DUE

        if self.last_phq9_score() and self.last_phq9_score() >= 5:
            result.add_narrative(f'{self.patient.first_name} had a PHQ9 score >= 5 '
                                 f'{self.display_date(self.last_phq9_date())}.')

        if self.last_gad7_score() and self.last_gad7_score() >= 5:
            result.add_narrative(f'{self.patient.first_name} had a GAD7 score >= 5 '
                                 f'{self.display_date(self.last_gad7_date())}.')

        result.add_recommendation(
            StructuredAssessmentRecommendation(
                key='CUSTOMER_SCAN_RECOMMEND_MOOD_DISORDER_STRUCTURED_ASSESSMENT',
                rank=1,
                button='Assess',
                patient=self.patient,
                questionnaires=[StructuredAssessmentQuestionnaireMoodDisorder],
                title='Assess Mood Disorder'))

        return result



from typing import List

from canvas_workflow_kit import events
from canvas_workflow_kit.protocol import ClinicalQualityMeasure
from canvas_workflow_kit.recommendation import (InterviewRecommendation, InstructionRecommendation,
                                PrescribeRecommendation, Recommendation)
# flake8: noqa
from canvas_workflow_kit.value_set.v2018 import (
    AnnualWellnessVisit, Ethnicity, FaceToFaceInteraction, HealthAndBehavioralAssessmentInitial,
    HealthAndBehavioralAssessmentReassessment, HealthBehavioralAssessmentIndividual,
    HomeHealthcareServices, LimitedLifeExpectancy, OccupationalTherapyEvaluation, OfficeVisit,
    OncAdministrativeSex, OphthalmologicalServices, Payer,
    PreventiveCareServicesEstablishedOfficeVisit18AndUp, PreventiveCareServicesGroupCounseling,
    PreventiveCareServicesIndividualCounseling, PreventiveCareServicesInitialOfficeVisit18AndUp,
    PreventiveCareServicesOther, PsychVisitDiagnosticEvaluation, PsychVisitPsychotherapy,
    Psychoanalysis, Race, SpeechAndHearingEvaluation, TobaccoNonUser,
    TobaccoUseCessationCounseling, TobaccoUseCessationPharmacotherapy, TobaccoUseScreening,
    TobaccoUser)


class ClinicalQualityMeasure138v6(ClinicalQualityMeasure):
    """
    Preventive Care and Screening: Tobacco Use: Screening and Cessation Intervention

    Description: Percentage of patients aged 18 years and older who were screened for tobacco use
    one or more times within 24 months AND who received tobacco cessation intervention if
    identified as a tobacco user

    Three rates are reported:
    a. Percentage of patients aged 18 years and older who were screened for tobacco use one or more
    times within 24 months
    b. Percentage of patients aged 18 years and older who were screened for tobacco use and
    identified as a tobacco user who received tobacco cessation intervention
    c. Percentage of patients aged 18 years and older who were screened for tobacco use one or more
    times within 24 months AND who received tobacco cessation intervention if identified as a
    tobacco user

    Definition: Tobacco Use - Includes any type of tobacco
    Tobacco Cessation Intervention - Includes brief counseling (3 minutes or less), and/or
    pharmacotherapy -- Note:  Concepts aligned with brief counseling (eg, minimal and intensive
    advice/counseling interventions conducted both in person and over the phone) are included in
    the value set for the numerator.  Other concepts such as written self-help materials (eg,
    brochures, pamphlets) and complementary/alternative therapies are not included in the value set
    and do not qualify for the numerator.

    Rationale: This measure is intended to promote adult tobacco screening and tobacco cessation
    interventions for those who use tobacco products. There is good evidence that tobacco screening
    and brief cessation intervention (including counseling and/or pharmacotherapy) is successful in
    helping tobacco users quit. Tobacco users who are able to stop using tobacco lower their risk
    for heart disease, lung disease, and stroke.

    Guidance: If a patient uses any type of tobacco (ie, smokes or uses smokeless tobacco), the
    expectation is that they should receive tobacco cessation intervention: either counseling
    and/or pharmacotherapy.

    If a patient has multiple tobacco use screenings during the 24 month period, only the most
    recent screening, which has a documented status of tobacco user or tobacco non-user, will be
    used to satisfy the measure requirements.

    If tobacco use status of a patient is unknown, the patient does not meet the screening
    component required to be counted in the numerator and should be considered a measure failure.
    Instances where tobacco use status of "unknown" is recorded include: 1) the patient was not
    screened; or 2) the patient was screened and the patient (or caregiver) was unable to provide a
    definitive answer.  If the patient does not meet the screening component of the numerator but
    has an allowable medical exception, then the patient should be removed from the denominator of
    the measure and reported as a valid exception.

    The medical reason exception may be applied to either the screening data element OR to any of
    the applicable tobacco cessation intervention data elements (counseling and/or pharmacotherapy)
    included in the measure.

    If a patient has a diagnosis of limited life expectancy, that patient has a valid denominator
    exception for not being screened for tobacco use or for not receiving tobacco use cessation
    intervention (counseling and/or pharmacotherapy) if identified as a tobacco user.

    As noted above in a recommendation statement from the USPSTF, the current evidence is
    insufficient to recommend electronic nicotine delivery systems (ENDS) including electronic
    cigarettes for tobacco cessation.  Additionally, ENDS are not currently classified as tobacco
    in the recent evidence review to support the update of the USPSTF recommendation given that the
    devices do not burn or use tobacco leaves.  In light of the current lack of evidence, the
    measure does not currently capture e-cigarette usage as either tobacco use or a cessation aid.

    The requirement of "Count >=2 Encounter, Performed" is to establish that the eligible
    professional or eligible clinician has an existing relationship with the patient for certain
    types of encounters.

    This measure contains three reporting rates which aim to identify patients who were screened
    for tobacco use (rate/population 1), patients who were identified as tobacco users and who
    received tobacco cessation intervention (rate/population 2), and a comprehensive look at the
    overall performance on tobacco screening and cessation intervention (rate/population 3). By
    separating this measure into various reporting rates, the eligible professional or eligible
    clinician will be able to better ascertain where gaps in performance exist, and identify
    opportunities for improvement. The overall rate (rate/population 3) can be utilized to compare
    performance to prior published versions of this measure.

    More information: https://ecqi.healthit.gov/ecqm/measures/cms138v6
    """

    responds_to_event_types = [
        events.HEALTH_MAINTENANCE,
    ]

    authors = [
        'American Medical Association (AMA)',
        'PCPI(R) Foundation (PCPI[R])',
    ]

    references = [
        'Siu AL; U.S. Preventive Services Task Force. Behavioral and Pharmacotherapy Interventions for Tobacco Smoking Cessation in Adults, Including Pregnant Women: U.S. Preventive Services Task Force Recommendation Statement. Ann Intern Med. 2015 Oct 20;163(8):622-34.',
    ]

    funding_source = ''

    def in_initial_population(self):
        """
        Initial population: All patients aged 18 years and older seen for at least two visits or at
        least one preventive visit during the measurement period
        """
        age = self.patient.age_at(self.timeframe.end)
        if age < 18 or age >= 76:
            return False

        # FIX: how do we count visits and tell the difference between a normal vist and a "preventive" visit?

        return True

    def in_denominator(self, population=1):
        """
        Denominator: Population 1:
        Equals Initial Population

        Population 2:
        Equals Initial Population who were screened for tobacco use and identified as a tobacco
        user

        Population 3:
        Equals Initial Population

        Exclusions: None

        Exceptions: Population 1:
        Documentation of medical reason(s) for not screening for tobacco use (eg, limited life
        expectancy, other medical reason)

        Population 2:
        Documentation of medical reason(s) for not providing tobacco cessation intervention (eg,
        limited life expectancy, other medical reason)

        Population 3:
        Documentation of medical reason(s) for not screening for tobacco use OR for not providing
        tobacco cessation intervention for patients identified as tobacco users (eg, limited life
        expectancy, other medical reason)
        """
        # oddly this protocol is actually three protocols packed into
        # one with three different denominator and numerator
        # definitions.  I don't know how we'd support this from a
        # caller perspective, but for now we'll take an optional param
        # to in_denominator and in_numerator which defaults to #1.
        if not self.in_initial_population():
            return False

        if self.patient.conditions.before(self.timeframe.start).find(LimitedLifeExpectancy):
            return False

        if population in (1, 3):
            return True
        elif population == 2 and self.patient.conditions.before(
                self.timeframe.start).find(TobaccoUser):
            return True

        return False

    def in_numerator(self, population=1):
        """
        Numerator: Population 1:
        Patients who were screened for tobacco use at least once within 24 months

        Population 2:
        Patients who received tobacco cessation intervention

        Population 3:
        Patients who were screened for tobacco use at least once within 24 months AND who received
        tobacco cessation intervention if identified as a tobacco user

        Exclusions: Not Applicable
        """
        screening = self.patient.interviews.within_two_years(
            self.timeframe).find(TobaccoUseScreening)
        meds = self.patient.medications.within_two_years(
            self.timeframe).find(TobaccoUseCessationPharmacotherapy)
        counseling = self.patient.instructions.within_two_years(
            self.timeframe).find(TobaccoUseCessationCounseling)

        if population == 1:
            if screening:
                return True
        elif population == 2:
            if counseling:
                return True
        elif population == 3:
            if screening and (meds or counseling):
                return True

        return False

    def compute_results(self):
        """
        Clinical recommendation: The USPSTF recommends that clinicians ask all adults about tobacco
        use, advise them to stop using tobacco, and provide behavioral interventions and U.S. Food
        and Drug Administration (FDA)-approved pharmacotherapy for cessation to adults who use
        tobacco. (Grade A Recommendation) (U.S. Preventive Services Task Force, 2015)

        The USPSTF recommends that clinicians ask all pregnant women about tobacco use, advise them
        to stop using tobacco, and provide behavioral interventions for cessation to pregnant women
        who use tobacco. (Grade A Recommendation) (U.S. Preventive Services Task Force, 2015)

        The USPSTF concludes that the current evidence is insufficient to recommend electronic
        nicotine delivery systems for tobacco cessation in adults, including pregnant women. The
        USPSTF recommends that clinicians direct patients who smoke tobacco to other cessation
        interventions with established effectiveness and safety (previously stated). (Grade I
        Statement) (U.S. Preventive Services Task Force, 2015)
        """
        patient = self.patient

        recommendations: List[Recommendation] = []
        status = 'not_applicable'
        narrative = ''

        # no screening in 2 years?  they should be screened
        screening = patient.interviews.within_two_years(
            self.timeframe).find(TobaccoUseScreening).last()
        if not screening:
            rec = InterviewRecommendation(patient=patient, questionnaire=TobaccoUseScreening)
            recommendations.append(rec)
            status = 'due'
            narrative = rec.narrative

        else:
            # they have been screened, see if they're a smoker
            # FIX: what's the right way to do this that doesn't depend on the particular language used?
            if 'currently' in screening['results'][0]['narrative'].lower():
                # advise meds or
                meds = patient.medications.within_two_years(
                    self.timeframe).find(TobaccoUseCessationPharmacotherapy)
                counseling = patient.instructions.within_two_years(
                    self.timeframe).find(TobaccoUseCessationCounseling)
                if meds or counseling:
                    status = 'current'
                    narrative = f'{patient.first_name} is a tobacco user but has received tobacco use cessation intervention.'
                else:
                    recommendations.append(
                        InstructionRecommendation(
                            patient=patient, instruction=TobaccoUseCessationCounseling))

                    recommendations.append(
                        PrescribeRecommendation(
                            patient=patient, prescription=TobaccoUseCessationPharmacotherapy))

                    status = 'due'
                    narrative = f'{patient.first_name} is a tobacco user and should receive tobacco use cessation intervention.'

        return {'status': status, 'narrative': narrative, 'recommendations': recommendations}



# type: ignore
from typing import Dict, List, Optional, cast

import arrow

from cached_property import cached_property

from canvas_workflow_kit import events, settings
from canvas_workflow_kit.protocol import (
    STATUS_DUE,
    STATUS_SATISFIED,
    ClinicalQualityMeasure,
    CodingStruct,
    ExternallyAwareClinicalQualityMeasure,
    ProtocolResult
)
from canvas_workflow_kit.recommendation import InterviewRecommendation
from canvas_workflow_kit.timeframe import Timeframe
from canvas_workflow_kit.value_set.customer_value_sets.scan import (
    QuestionnaireHealthCHECReferrals,
    QuestionnairePVQ
)


class CustomerScanPhysicalActivityIntervention(ExternallyAwareClinicalQualityMeasure,
                                               ClinicalQualityMeasure):
    """
    Look out for patients reporting rare or light physical activity and recommend the
    physical activity intervention if needed.
    """

    class Meta:
        title = 'SCAN: Physical Activity Intervention'
        version = '2020-10-13v1'
        description = ''
        information = 'https://docs.google.com/document/d/1NW4b3OT_yPeZy_oIKt_DCCeQ0egDaWFE/edit'

        identifiers = ['CustomerScanPhysicalActivityInterventionV1']

        types = ['CQM']

        responds_to_event_types = [
            events.HEALTH_MAINTENANCE,
        ]

        authors = [
            'SCAN',
        ]

        compute_on_change_types = [
            ClinicalQualityMeasure.CHANGE_PROTOCOL_OVERRIDE,
            ClinicalQualityMeasure.CHANGE_INTERVIEW,
        ]

        funding_source = 'SCAN'

        references = ['Written by SCAN']

    @classmethod
    def enabled(cls) -> bool:
        return cast(bool, settings.IS_HEALTHCHEC)

    def in_initial_population(self) -> bool:
        return True

    def interview_lookback_timeframe(self) -> Timeframe:
        return Timeframe(start=self.timeframe.end.shift(days=-30), end=self.timeframe.end)

    @cached_property
    def recent_pvqs(self) -> List[Dict]:
        interviews = self.patient.interviews.find(QuestionnairePVQ).within(
            self.interview_lookback_timeframe())

        for interview in interviews:
            interview['groupedResponses'] = self.group_question_responses(interview)

        return interviews

    @cached_property
    def latest_how_physically_active_response(self) -> Optional[Dict]:
        how_physically_active_question_coding = CodingStruct(
            code='123978lijsdf169', system='INTERNAL')

        return self.latest_response_to_question(
            question_coding=how_physically_active_question_coding)

    def latest_how_physically_active_response_date(self) -> Optional[arrow.Arrow]:
        if self.latest_how_physically_active_response:
            return arrow.get(self.latest_how_physically_active_response.get('note_timestamp'))

    def latest_response_to_question(self, question_coding: CodingStruct) -> Optional[Dict]:
        for interview in sorted(
                self.recent_pvqs, key=lambda i: arrow.get(i['noteTimestamp']), reverse=True):
            question_responses = interview.get('groupedResponses', {}).get(
                question_coding.code, [])
            if question_responses:
                return {
                    'note_timestamp': interview['noteTimestamp'],
                    'question_coding': question_coding,
                    'responses': question_responses,
                }

    @cached_property
    def last_referral_for_physical_activity_intervention(self) -> Optional[Dict]:
        interviews = self.patient.interviews.find(QuestionnaireHealthCHECReferrals).within(
            self.interview_lookback_timeframe())

        if not interviews:
            return None

        # Go through them in order with the most recent one first and working
        # backwards
        # return the first one with a response representing a physical
        # activity intervention
        for interview in sorted(
                interviews, key=lambda i: arrow.get(i['noteTimestamp']), reverse=True):
            for response in interview.get('responses', []):
                if response.get(
                        'code'
                ) == '390893007' and 'Physical Activity Intervention' in response.get('value', ''):
                    return interview

        # nothing found
        return None

    def last_referral_for_physical_activity_intervention_date(self) -> Optional[arrow.Arrow]:
        if self.last_referral_for_physical_activity_intervention:
            return arrow.get(
                self.last_referral_for_physical_activity_intervention['noteTimestamp'])
        return None

    def referred_for_physical_activity_intervention_on_or_after(
            self, reference_date: arrow.Arrow) -> bool:
        # It can't have been referred after a date that doesn't exist!
        if not reference_date:
            return False

        # If it hasn't been referred, it certainly wasn't referred after the
        # reference date.
        if not self.last_referral_for_physical_activity_intervention_date():
            return False

        return self.last_referral_for_physical_activity_intervention_date() >= reference_date

    def in_denominator(self) -> bool:
        """
        Denominator: Equals Initial Population

        Exclusions: Exclude patients who were in hospice care during the measurement year

        Exceptions: None
        """
        if not self.in_initial_population():
            return False

        if self.patient.hospice_within(self.timeframe):
            return False

        return True

    def in_numerator(self) -> bool:
        """
        Patient is in the numerator if they have not indicated inadequate physical activity OR if
        they have indicated inadequate physical activity, but they have been referred for a
        physical activity intervention since then.
        """

        inadequate_physical_activity_reported = False

        rarely_physically_active_answer_coding = CodingStruct(
            code='123978lijsdf170', system='INTERNAL')
        light_not_every_week_physically_active_answer_coding = CodingStruct(
            code='123978lijsdf171', system='INTERNAL')
        light_every_week_physically_active_answer_coding = CodingStruct(
            code='123978lijsdf172', system='INTERNAL')

        if self.latest_how_physically_active_response:
            set_of_answers_indicating_additional_monitoring_required = set([
                rarely_physically_active_answer_coding.code,
                light_not_every_week_physically_active_answer_coding.code,
                light_every_week_physically_active_answer_coding.code
            ])
            set_of_actual_responses = set(
                self.latest_how_physically_active_response.get('responses', []))
            inadequate_physical_activity_reported = len(
                set_of_actual_responses.intersection(
                    set_of_answers_indicating_additional_monitoring_required)) > 0

        no_inadequate_physical_activity_reported = not inadequate_physical_activity_reported

        referred_for_physical_activity_intervention = self.referred_for_physical_activity_intervention_on_or_after(  # noqa: E501
            self.latest_how_physically_active_response_date())

        return no_inadequate_physical_activity_reported or (
            inadequate_physical_activity_reported and referred_for_physical_activity_intervention)

    def craft_satisfied_result(self):
        """
        Satisfied if in the numerator. (See in_numerator description for inclusion criteria)
        """
        result = ProtocolResult()

        result.due_in = -1
        result.status = STATUS_SATISFIED

        referred_for_physical_activity_intervention = self.referred_for_physical_activity_intervention_on_or_after(  # noqa: E501
            self.latest_how_physically_active_response_date())

        if referred_for_physical_activity_intervention:
            result.add_narrative(
                f'{self.patient.first_name} reported suboptimal physical '
                'activity, but was referred for physical activity intervention '
                f'{self.display_date(self.last_referral_for_physical_activity_intervention_date())}.'  # noqa: E501
            )
        else:
            result.add_narrative(f'{self.patient.first_name} has not reported suboptimal physical '
                                 'activity in the past 30 days.')

        return result

    def craft_unsatisfied_result(self):
        """
        Unsatisfied if not in the numerator. Recommend a Physical Activity Intervention.
        """
        result = ProtocolResult()

        result.due_in = -1
        result.status = STATUS_DUE

        result.add_narrative(
            f'{self.patient.first_name} reported suboptimal physical activity '
            f'{self.display_date(self.latest_how_physically_active_response_date())}.')

        result.add_recommendation(
            InterviewRecommendation(
                key='CUSTOMER_SCAN_RECOMMEND_PHYSICAL_ACTIVITY_INTERVENTION_REFERRAL',
                rank=1,
                button='Refer',
                patient=self.patient,
                questionnaires=[QuestionnaireHealthCHECReferrals],
                title='Refer for Physical Activity Intervention'))

        return result



class CanvasCodeSet:
    CODES = {
        'CANVAS0001': {
            'code': 'CANVAS0001',
            'description': 'Questionnaire used to determine if the patient has sufficient risk for COVID-19 to require a follow-up.',
            'short_description': 'COVID-19 Follow-up qualification questionnaire',
        },
        'CANVAS0002': {
            'code': 'CANVAS0002',
            'description': 'Question used to determine if the patient has exhibited symptoms commonly associated with COVID-19.',
            'short_description': 'COVID-19 Symptoms Question',
        },
        'CANVAS0003': {
            'code': 'CANVAS0003',
            'description': 'Question used to determine if the patient has traveled within the past 14 days.',
            'short_description': 'COVID-19 Recent Travel Question',
        },
        'CANVAS0004': {
            'code': 'CANVAS0004',
            'description': 'Question used to determine if the patient has been exposed to someone with a confirmed case of COVID-19.',
            'short_description': 'COVID-19 Exposure Question',
        },
        'CANVAS0005': {
            'code': 'CANVAS0005',
            'description': 'Question used to override COVID-19 risk scoring and force a follow-up.',
            'short_description': 'COVID-19 Forced Follow-up Question',
        },
        'CANVAS0006': {
            'code': 'CANVAS0006',
            'description': 'Questionnaire used to check in on patients with higher risk for COVID-19',
            'short_description': 'COVID-19 High Risk Check-in Questionnaire',
        },
        'CANVAS0007': {
            'code': 'CANVAS0007',
            'description': "Question used to gauge patient's knowledge of the Coronavirus situation.",
            'short_description': 'COVID-19 Knowledge Check Question',
        },
        'CANVAS0008': {
            'code': 'CANVAS0008',
            'description': 'Question used to check which COVID-19 prevention steps the patient has started using.',
            'short_description': 'COVID-19 Preventive Steps Employed Question',
        },
        'CANVAS0009': {
            'code': 'CANVAS0009',
            'description': 'Question used to ask if the patient has any questions about COVID-19 prevention.',
            'short_description': 'Question About COVID-19 Prevention Questions',
        },
        'CANVAS0010': {
            'code': 'CANVAS0010',
            'description': 'Question used to determine patient worry about COVID-19 impacts.',
            'short_description': 'COVID-19 Impact Worry Question',
        },
        'CANVAS0011': {
            'code': 'CANVAS0011',
            'description': 'Question used to identify patient social isolation mitigation strategy.',
            'short_description': 'COVID-19 Social Isolation Mitigation Strategy Question',
        },
        'CANVAS0012': {
            'code': 'CANVAS0012',
            'description': 'Question asking what help the patient would like from the clinic.',
            'short_description': 'Question asking what help the patient would like from the clinic.',
        },
        'CANVAS0013': {
            'code': 'CANVAS0013',
            'description': 'Question used to note the internal assessment of the COVID-19 outreach call.',
            'short_description': 'COVID-19 Outreach Call Internal Assessment',
        },
    }

    @staticmethod
    def code(key: str) -> str:
        return CanvasCodeSet.CODES.get(key, {'code': ''})['code']



from protocols import events
from protocols.protocol import ClinicalQualityMeasure
from protocols.value_set.v2021 import (
    AdolescentDepressionMedications, AdultDepressionMedications, BipolarDiagnosis,
    DepressionDiagnosis, EncounterToScreenForDepression, Ethnicity,
    FollowUpForAdolescentDepression, FollowUpForAdultDepression, OncAdministrativeSex, Payer,
    PhysicalTherapyEvaluation, Race, ReferralForAdolescentDepression, ReferralForAdultDepression)


class ClinicalQualityMeasure2v10(ClinicalQualityMeasure):
    """
    Preventive Care and Screening: Screening for Depression and Follow-Up Plan

    Description: Percentage of patients aged 12 years and older screened for depression on the date
    of the encounter or up to 14 days prior to the date of the encounter using an age-appropriate
    standardized depression screening tool AND if positive, a follow-up plan is documented on the
    date of the eligible encounter

    Definition: Screening:
    Completion of a clinical or diagnostic tool used to identify people at risk of developing or
    having a certain disease or condition, even in the absence of symptoms.

    Standardized Depression Screening Tool: A normalized and validated depression screening tool
    developed for the patient population in which it is being utilized.

    Examples of standardized depression screening tools include but are not limited to:
    *  Adolescent Screening Tools (12-17 years)
       *  Patient Health Questionnaire for Adolescents (PHQ-A)
       *  Beck Depression Inventory-Primary Care Version (BDI-PC)
       *  Mood Feeling Questionnaire (MFQ)
       *  Center for Epidemiologic Studies Depression Scale (CES-D)
       *  Patient Health Questionnaire (PHQ-9)
       *  Pediatric Symptom Checklist (PSC-17)
       *  PRIME MD-PHQ2
    *  Adult Screening Tools (18 years and older)
       *  Patient Health Questionnaire (PHQ9)
       *  Beck Depression Inventory (BDI or BDI-II)
       *  Center for Epidemiologic Studies Depression Scale (CES-D)
       *  Depression Scale (DEPS)
       *  Duke Anxiety-Depression Scale (DADS)
       *  Geriatric Depression Scale (GDS)
       *  Cornell Scale for Depression in Dementia (CSDD)
       *  PRIME MD-PHQ2
       *  Hamilton Rating Scale for Depression (HAM-D)
       *  Quick Inventory of Depressive Symptomatology Self-Report (QID-SR)
       *  Computerized Adaptive Testing Depression Inventory (CAT-DI)
       *  Computerized Adaptive Diagnostic Screener (CAD-MDD)
    * Perinatal Screening Tools
       *  Edinburgh Postnatal Depression Scale
       *  Postpartum Depression Screening Scale
       *  Patient Health Questionnaire 9 (PHQ-9)
       *  Beck Depression Inventory
       *  Beck Depression Inventory-II
       *  Center for Epidemiologic Studies Depression Scale
       *  Zung Self-rating Depression Scale

    Follow-Up Plan:
    Documented follow-up for a positive depression screening must include one or more of the
    following:
     *  Referral to a practitioner who is qualified to diagnose and treat depression
     *  Pharmacological interventions
     *  Other interventions or follow-up for the diagnosis or treatment of depression

    Rationale: Depression is a serious medical illness associated with higher rates of chronic
    disease, increased health care utilization, and impaired functioning (Katon, 2003; Wells et
    al., 1989). 2016 U.S. survey data indicate that 12.8 percent of adolescents (3.1 million
    adolescents) had a major depressive episode (MDE) in the past year, with nine percent of
    adolescents (2.2 million adolescents) having one MDE with severe impairment. The same data
    indicate that 6.7 percent of adults aged 18 or older (16.2 million adults) had at least one MDE
    with 4.3 percent of adults (10.3 million adults) having one MDE with severe impairment in the
    past year (Substance Abuse and Mental Health Services Administration, 2017). Data indicate that
    severity of depressive symptoms factor into having difficulty with work, home, or social
    activities. For example, as the severity of depressive symptoms increased, rates of having
    difficulty with work, home, or social activities related to depressive symptoms increased. For
    those twelve and older with mild depressive symptoms, 45.7% reported difficulty with activities
    and those with severe depressive symptoms, 88.0% reported difficulty (Pratt & Brody, 2014).
    Children and teens with major depressive disorder (MDD) have been found to have difficulty
    carrying out their daily activities, relating to others, growing up healthy, and also are at an
    increased risk of suicide (Siu on behalf of the U.S. Preventive Services Task Force [USPSTF],
    2016). Additionally, perinatal depression (considered here as depression arising in the period
    from conception to the end of the first postnatal year) affects up to 12% of women (Woody,
    Ferrari, Siskind, Whiteford, & Harris, 2017). Depression and other mood disorders, such as
    bipolar disorder and anxiety disorders, especially during the perinatal period, can have
    devastating effects on women, infants, and families (American College of Obstetricians and
    Gynecologists, 2018). Maternal suicide rates rise over hemorrhage and hypertensive disorders as
    a cause of maternal mortality (Palladino, Singh, Campbell, Flynn, & Gold, 2011).

    Negative outcomes associated with depression make it crucial to screen in order to identify and
    treat depression in its early stages. While Primary Care Providers (PCPs) serve as the first
    line of defense in the detection of depression, studies show that PCPs fail to recognize up to
    50% of depressed patients (Borner, Braunstein, St. Victor, & Pollack, 2010). "In nationally
    representative U.S. surveys, about eight percent of adolescents reported having major
    depression in the past year. Only 36% to 44% of children and adolescents with depression
    receive treatment, suggesting that the majority of depressed youth are undiagnosed and
    untreated" (Siu on behalf of USPSTF, 2016, p. 360 & p. 364). Evidence supports that screening
    for depression in pregnant and postpartum women is of moderate net benefit and treatment
    options for positive depression screening should be available for patients twelve and older
    including pregnant and postpartum women.

    If preventing negative patient outcomes is not enough, the substantial economic burden of
    depression for individuals and society alike makes a case for screening for depression on a
    regular basis. Depression imposes economic burden through direct and indirect costs: "In the
    United States, an estimated $22.8 billion was spent on depression treatment in 2009, and lost
    productivity cost an additional estimated $23 billion in 2011" (Siu & USPSTF, 2016, p.
    383-384).

    This measure seeks to align with clinical guideline recommendations as well as the Healthy
    People 2020 recommendation for routine screening for mental health problems as a part of
    primary care for both children and adults (U.S. Department of Health and Human Services, 2014)
    and makes an important contribution to the quality domain of community and population health.

    Guidance: A depression screen is completed on the date of the encounter or up to 14 days prior
    to the date of the encounter using an age-appropriate standardized depression screening tool
    AND if positive, a follow-up plan must be documented on the date of the encounter, such as
    referral to a practitioner who is qualified to treat depression, pharmacological interventions
    or other interventions for the treatment of depression.

    This eCQM is a patient-based measure. Depression screening is required once per measurement
    period, not at all encounters.

    The intent of the measure is to screen for depression in patients who have never had a
    diagnosis of depression or bipolar disorder prior to the eligible encounter used to evaluate
    the numerator. Patients who have ever been diagnosed with depression or bipolar disorder will
    be excluded from the measure.

    Screening Tools:
     *  An age-appropriate, standardized, and validated depression screening tool must be used for
    numerator compliance.
     *  The name of the age-appropriate standardized depression screening tool utilized must be
    documented in the medical record.
     *  The depression screening must be reviewed and addressed in the office of the provider,
    filing the code, on the date of the encounter. Positive pre-screening results indicating a
    patient is at high risk for self-harm should receive more urgent intervention as determined by
    the provider practice.
     *  The screening should occur during a qualifying encounter or up to 14 days prior to the date
    of the qualifying encounter.
     *  The measure assesses the most recent depression screening completed either during the
    eligible encounter or within the 14 days prior to that encounter. Therefore, a clinician would
    not be able to complete another screening at the time of the encounter to count towards a
    follow-up, because that would serve as the most recent screening. In order to satisfy the
    follow-up requirement for a patient screening positively, the eligible clinician would need to
    provide one of the aforementioned follow-up actions, which does not include use of a
    standardized depression screening tool.

    Follow-Up Plan:

    The follow-up plan must be related to a positive depression screening, for example: "Patient
    referred for psychiatric evaluation due to positive depression screening."

    Examples of a follow-up plan include but are not limited to:
     *  Referral to a practitioner or program for further evaluation for depression, for example,
    referral to a psychiatrist, psychologist, social worker, mental health counselor, or other
    mental health service such as family or group therapy, support group, depression management
    program, or other service for treatment of depression
     *  Other interventions designed to treat depression such as behavioral health evaluation,
    psychotherapy, pharmacological interventions, or additional treatment options

    Should a patient screen positive for depression, a clinician should opt to complete a suicide
    risk assessment when appropriate and based on individual patient characteristics. However, for
    the purposes of this measure, a suicide risk assessment or additional screening using a
    standardized tool will not qualify as a follow-up plan.

    This version of the eCQM uses QDM version 5.5. Please refer to the eCQI resource center
    (https://ecqi.healthit.gov/qdm) for more information on the QDM.

    More information: https://ecqi.healthit.gov/ecqm/measures/cms2v10
    """

    title = 'Preventive Care and Screening: Screening for Depression and Follow-Up Plan'

    identifiers = ['CMS2v10']

    types = ['CQM']

    responds_to_event_types = [
        events.HEALTH_MAINTENANCE,
    ]

    authors = [
        'PCPI(R) Foundation (PCPI[R])',
    ]

    references = [
        'American College of Obstetricians and Gynecologists, Committee on Obstetric Practice. (2018). ACOG Committee Opinion Number 757: Screening for perinatal depression. Obstetrics and Gynecology, 132(5), e208-e212. doi: 10.1097/AOG.0000000000002927',
        'Borner, I., Braunstein, J. W., St. Victor, R., & Pollack, J. (2010). Evaluation of a 2-question screening tool for detecting depression in adolescents in primary care. Clinical Pediatrics, 49(10), 947-995. doi:10.1177/0009922810370203',
        'Katon, W. J. (2003). Clinical and health services relationships between major depression, depressive symptoms, and general medical illness. Biological Psychiatry, 54(3), 216-226. doi: 10.1016/s0006-3223(03)00273-7',
        'Palladino, C. L., Singh, V., Campbell, J., Flynn, H., & Gold, K. (2011). Homicide and suicide during the perinatal period: Findings from the National Violent Death Reporting System. Obstetrics and Gynecology, 118(5), 1056-1063. doi:10.1097/AOG.0b013e31823294da',
        'Pratt, L. A., & Brody, D. J. (2014). Depression in the U.S. household population, 2009-2012. NCHS Data Brief No. 172. Hyattsville, MD: U.S. Department of Health and Human Services, Centers for Disease Control and Prevention, National Center for Health Statistics. Retrieved from https://www.cdc.gov/nchs/data/databriefs/db172.pdf',
        'Siu, A. L., on behalf of USPSTF. (2016). Screening for depression in children and adolescents: U.S. Preventive Services Task Force recommendation statement. Annals of Internal Medicine, 164(5), 360-366. doi:10.7326/M15-2957',
        'Siu, A. L., & USPSTF. (2016). Screening for depression in adults: U.S. Preventive Services Task Force recommendation statement. Journal of the American Medical Association, 315(4), 380-387. doi:10.1001/jama.2015.18392.',
        'Substance Abuse and Mental Health Services Administration. (2017). Key substance use and mental health indicators in the United States: Results from the 2016 National Survey on Drug Use and Health. Rockville, MD: Center for Behavioral Health Statistics and Quality, Substance Abuse and Mental Health Services Administration. Retrieved from https://www.samhsa.gov/data/sites/default/files/NSDUH-FFR1-2016/NSDUH-FFR1-2016.htm',
        'Trangle, M., Gursky, J., Haight, R., Hardwig, J., Hinnenkamp, T., Kessler, D.,… Myszkowski, M. (2016). Adult depression in primary care. Bloomington, MN: Institute for Clinical Systems Improvement. Retrieved from https://www.icsi.org/guideline/depression/',
        'U.S. Department of Health and Human Services. (2014). Healthy People 2020: Mental health and mental disorders. Washington, DC: U.S. Department of Health and Human Services. Retrieved from http://www.healthypeople.gov/2020/topicsobjectives2020/objectiveslist.aspx?topicId=28',
        'Wells, K. B., Stewart, A., Hays, R. D., Burnam, M. A., Rogers, W., Daniels, M., … Ware, J. (1989). The functioning and well-being of depressed patients. Results from the Medical Outcomes Study. Journal of the American Medical Association, 262(7), 914-919. Abstract retrieved from https://www.ncbi.nlm.nih.gov/pubmed/2754791',
        'Woody, C. A., Ferrari, A. J., Siskind, D. J., Whiteford, H. A., & Harris, M. G. (2017). A systematic review and meta-regression of the prevalence and incidence of perinatal depression. Journal of Affective Disorders, 219, 88-92. doi:10.1016/j.jad.2017.05.003',
    ]

    funding_source = ''

    def in_initial_population(self):
        """
        Initial population: All patients aged 12 years and older at the beginning of the
        measurement period with at least one eligible encounter during the measurement period
        """
        pass

    def in_denominator(self):
        """
        Denominator: Equals Initial Population

        Exclusions: Patients who have been diagnosed with depression or with bipolar disorder

        Exceptions: Patient Reason(s)
        Patient refuses to participate
        OR
        Medical Reason(s)
        Documentation of medical reason for not screening patient for depression (e.g., cognitive,
        functional, or motivational limitations that may impact accuracy of results; patient is in
        an urgent or emergent situation where time is of the essence and to delay treatment would
        jeopardize the patient's health status)
        """
        return self.in_initial_population()

    def in_numerator(self):
        """
        Numerator: Patients screened for depression on the date of the encounter or up to 14 days
        prior to the date of the encounter using an age-appropriate standardized tool AND if
        positive, a follow-up plan is documented on the date of the eligible encounter

        Exclusions: Not Applicable
        """
        pass


    def compute_results(self):
        """
        Clinical recommendation: Adolescent Recommendation (12-18 years):

        "The USPSTF recommends screening for MDD in adolescents aged 12 to 18 years. Screening
        should be implemented with adequate systems in place to ensure accurate diagnosis,
        effective treatment, and appropriate follow-up (B recommendation)" (Siu on behalf of
        USPSTF, 2016, p. 360).

        Adult Recommendation (18 years and older):

        "The USPSTF recommends screening for depression in the general adult population, including
        pregnant and postpartum women. Screening should be implemented with adequate systems in
        place to ensure accurate diagnosis, effective treatment, and appropriate follow-up (B
        recommendation)" (Siu & USPSTF, 2016, p. 380).

        The Institute for Clinical Systems Improvement (ICSI) health care guideline, Adult
        Depression in Primary Care, provides the following recommendations:
        1. "Clinicians should routinely screen all adults for depression using a standardized
        instrument."
        2. "Clinicians should establish and maintain follow-up with patients."
        3. "Clinicians should screen and monitor depression in pregnant and post-partum women."
        (Trangle et al., 2016, p. 8-10).
        """
        pass



import arrow

from canvas_workflow_kit import settings

from canvas_workflow_kit.builtin_cqms.hcc002v2_ckd_suspect import Hcc002v2
from canvas_workflow_kit.patient_recordset import ConditionRecordSet, ProtocolOverrideRecordSet
from canvas_workflow_kit.tests.base import SDKBaseTest
from canvas_workflow_kit.timeframe import Timeframe


class TestHcc002v2(SDKBaseTest):

    def setUp(self):
        self.mocks_path = f'{settings.BASE_DIR}/builtin_cqms/tests/mock_data'

    def test_enabled(self):
        self.assertTrue(Hcc002v2.enabled())

    def test_description(self):
        expected = ('All patients with evidence of two or more elevated eGFR values '
                    'and no active CKD problem on the Conditions List.')
        self.assertEqual(expected, Hcc002v2._meta.description)

    def test_information(self):
        expected = 'https://canvas-medical.zendesk.com/hc/en-us/articles/360059083713-CKD-Suspect-HCC002v2'
        self.assertEqual(expected, Hcc002v2._meta.information)

    def test_change_types(self):
        result = Hcc002v2._meta.compute_on_change_types
        expected = [
            'protocol_override',
            'condition',
            'lab_report',
            'patient',
        ]
        self.assertEqual(len(expected), len(result))
        for change in expected:
            self.assertIn(change, result)

    def test_high_creatine_levels(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        # patient has two lab reports with creatinine eGFR under 60 within the last 2 years
        patient = self.load_patient('hcc002v2_diagnosed')
        tested = Hcc002v2(patient=patient, timeframe=timeframe)
        self.assertEqual(2, len(tested.high_creatine_levels))

        # patient has two lab report within the last 2 years,
        # only one with creatinine eGFR under 60
        patient = self.load_patient('hcc002v2_diagnosed')
        patient.lab_reports[0]['value'] = '1.1'
        tested = Hcc002v2(patient=patient, timeframe=timeframe)
        self.assertEqual(1, len(tested.high_creatine_levels))

        # patient has two lab reports with creatinine eGFR under 60,
        # only one within the last 2 years
        patient = self.load_patient('hcc002v2_diagnosed')
        patient.lab_reports[0]['originalDate'] = '2016-08-22'
        tested = Hcc002v2(patient=patient, timeframe=timeframe)
        self.assertEqual(1, len(tested.high_creatine_levels))

    def test_high_creatine_levels_with_override(self):
        patient = self.load_patient('hcc002v2_diagnosed')
        patient.protocol_overrides = ProtocolOverrideRecordSet([{
            'protocolKey': 'Hcc002v2',
            'adjustment': {
                'reference': '2018-02-01T00:00:00Z',
                'cycleDays': 180
            },
            'snooze': None,
            'modified': '2018-03-15T12:32:54Z',
        }])  # yapf: disable
        patient.lab_reports[0]['originalDate'] = '2018-08-01'
        patient.lab_reports[1]['originalDate'] = '2018-09-01'

        tested = Hcc002v2(patient=patient, now=arrow.get('2018-08-01').shift(days=120))
        self.assertEqual(2, len(tested.high_creatine_levels))
        tested = Hcc002v2(patient=patient, now=arrow.get('2018-08-01').shift(days=180))
        self.assertEqual(2, len(tested.high_creatine_levels))
        tested = Hcc002v2(patient=patient, now=arrow.get('2018-08-01').shift(days=181))
        self.assertEqual(1, len(tested.high_creatine_levels))

    def test_has_active_condition(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        # patient has an active condition in HCC
        patient = self.load_patient('hcc002v2_diagnosed')
        tested = Hcc002v2(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.has_active_condition)

        # patient has NO active condition in HCC
        patient = self.load_patient('hcc002v2_diagnosed')
        patient.conditions[0]['clinicalStatus'] = 'XXXXXXXX'
        tested = Hcc002v2(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.has_active_condition)

        # patient has a condition NOT in HCC
        patient = self.load_patient('hcc002v2_diagnosed')
        patient.conditions[0]['coding'][0]['code'] = 'XXXXXXXX'
        patient.conditions[0]['coding'][1]['code'] = 'XXXXXXXX'
        tested = Hcc002v2(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.has_active_condition)

        # patient has NO condition
        patient = self.load_patient('hcc002v2_diagnosed')
        patient.conditions = ConditionRecordSet([])
        tested = Hcc002v2(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.has_active_condition)

    def test_eGFR(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        now = arrow.get('2018-12-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        patient = self.load_patient('hcc002v2_diagnosed')
        tested = Hcc002v2(patient=patient, timeframe=timeframe, now=now)

        # change of units
        result = tested.eGFR(1.37, 'mg/dL')
        expected = 66.45
        self.assertAlmostEqual(expected, result, 2)

        result = tested.eGFR(1.37 * 88.4, 'µmol/L')
        expected = 66.45
        self.assertAlmostEqual(expected, result, 2)

        # sex influence
        tested.patient.patient['sexAtBirth'] = 'F'
        result = tested.eGFR(1.37, 'mg/dL')
        expected = 49.31
        self.assertAlmostEqual(expected, result, 2)
        tested.patient.patient['sexAtBirth'] = 'M'

        # race influence
        tested.patient.patient['biologicalRaceCode'] = '2106-3'
        result = tested.eGFR(1.37, 'mg/dL')
        expected = 54.92
        self.assertAlmostEqual(expected, result, 2)
        tested.patient.patient['biologicalRaceCode'] = '2054-5'

        # age influence
        tested.patient.patient['birthDate'] = '1941-01-01'
        result = tested.eGFR(1.37, 'mg/dL')
        expected = 64.80
        self.assertAlmostEqual(expected, result, 2)
        tested.patient.patient['birthDate'] = '1951-01-01'

    def test_in_initial_population(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        # any patient --> true
        patient = self.load_patient('hcc002v2_diagnosed')
        tested = Hcc002v2(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_initial_population())

    def test_in_denominator(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        # patient has two lab reports with creatinine above 1.1 within the last 2 years
        patient = self.load_patient('hcc002v2_diagnosed')
        tested = Hcc002v2(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_denominator())

        # patient has two lab report within the last 2 years, only one with creatinine above 1.1
        patient = self.load_patient('hcc002v2_diagnosed')
        patient.lab_reports[0]['value'] = '1.1'
        tested = Hcc002v2(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_denominator())

        # patient has two lab reports with creatinine above 1.1, only one within the last 2 years
        patient = self.load_patient('hcc002v2_diagnosed')
        patient.lab_reports[0]['originalDate'] = '2016-08-22'
        tested = Hcc002v2(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_denominator())

    def test_in_numerator(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        # patient has a active condition in HCC
        patient = self.load_patient('hcc002v2_diagnosed')
        tested = Hcc002v2(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_numerator())

        # patient has NO active condition in HCC
        patient = self.load_patient('hcc002v2_diagnosed')
        patient.conditions[0]['clinicalStatus'] = 'XXXXXXXX'
        tested = Hcc002v2(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_numerator())

        # patient has a condition NOT in HCC
        patient = self.load_patient('hcc002v2_diagnosed')
        patient.conditions[0]['coding'][0]['code'] = 'XXXXXXXX'
        patient.conditions[0]['coding'][1]['code'] = 'XXXXXXXX'
        tested = Hcc002v2(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_numerator())

        # patient has NO condition
        patient = self.load_patient('hcc002v2_diagnosed')
        patient.conditions = ConditionRecordSet([])
        tested = Hcc002v2(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_numerator())

    def test_compute_results(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        # patient has a active condition in HCC and two lab reports
        # with creatinine above 1.1 within the last 2 years
        # --> satisfied
        patient = self.load_patient('hcc002v2_diagnosed')
        tested = Hcc002v2(patient=patient, timeframe=timeframe)
        result = tested.compute_results()
        self.assertEqual('satisfied', result.status)
        text = ''
        self.assertEqual(text, result.narrative)
        self.assertEqual([], result.recommendations)
        self.assertIsNone(result.due_in)

        # patient has NO active condition in HCC and two lab reports
        # with creatinine above 1.1 within the last 2 years
        # --> due
        patient = self.load_patient('hcc002v2_diagnosed')
        tested = Hcc002v2(patient=patient, timeframe=timeframe)
        patient.conditions[0]['coding'][0]['code'] = 'XXXXXXXX'
        patient.conditions[0]['coding'][1]['code'] = 'XXXXXXXX'
        result = tested.compute_results()
        self.assertEqual('due', result.status)
        text = (f'Jim has at least two eGFR measurements < 60 ml/min '
                f'over the last two years suggesting renal disease. '
                f'There is no associated condition on the Conditions List.')
        self.assertEqual(text, result.narrative)
        expected = [
            {
                'title': (f'Consider updating the Conditions List to include kidney '
                          f'related problems as clinically appropriate'),
                'command': {
                    'key': 'diagnose'
                },
                'key': 'HCC002v2_RECOMMEND_DIAGNOSE',
                'rank': 1,
                'button': 'Diagnose',
            },
        ]  # yapf: disable
        self.helper_exam_recommended(expected, result.recommendations)
        self.assertEqual(-1, result.due_in)

        # patient has one lab report with creatinine above 1.1 within the last 2 years
        # --> n/a
        patient = self.load_patient('hcc002v2_diagnosed')
        patient.conditions = ConditionRecordSet([])
        patient.lab_reports[0]['value'] = '1.1'
        tested = Hcc002v2(patient=patient, timeframe=timeframe)
        result = tested.compute_results()
        self.assertIsNotApplicable(result)

    def test_compute_results_with_override(self):
        patient = self.load_patient('hcc002v2_diagnosed')
        patient.protocol_overrides = ProtocolOverrideRecordSet([{
            'protocolKey': 'Hcc002v2',
            'adjustment': {
                'reference': '2018-02-01T00:00:00Z',
                'cycleDays': 180
            },
            'snooze': None,
            'modified': '2018-03-15T12:32:54Z',
        }])  # yapf: disable
        patient.lab_reports[0]['originalDate'] = '2018-08-01'
        patient.lab_reports[1]['originalDate'] = '2018-09-01'
        patient.conditions[0]['coding'][0]['code'] = 'XXXXXXXX'
        patient.conditions[0]['coding'][1]['code'] = 'XXXXXXXX'

        tested = Hcc002v2(patient=patient, now=arrow.get('2018-08-01').shift(days=120))
        result = tested.compute_results()
        self.assertEqual('due', result.status)
        self.assertEqual(-1, result.due_in)

        tested = Hcc002v2(patient=patient, now=arrow.get('2018-08-01').shift(days=180))
        result = tested.compute_results()
        self.assertEqual('due', result.status)
        self.assertEqual(-1, result.due_in)

        tested = Hcc002v2(patient=patient, now=arrow.get('2018-08-01').shift(days=181))
        result = tested.compute_results()
        self.assertIsNotApplicable(result)

    def helper_exam_recommended(self, expected, result):
        for exp, recommendation in zip(expected, result):
            self.assertEqual(exp['key'], recommendation.key)
            self.assertEqual(exp['rank'], recommendation.rank)
            self.assertEqual(exp['button'], recommendation.button)
            self.assertEqual(exp['title'], recommendation.title)
            self.assertEqual(exp['command']['key'], recommendation.command['key'])
